r"""
:class:`Holding`\s describe :class:`.Opinion`\s` attitudes toward :class:`.Rule`\s.

:class:`Holding`\s are text passages within :class:`.Opinion`\s
in which :class:`.Court` posits, or rejects, the validity of a
:class:`.Rule` within the :class:`.Court`\'s :class:`.Jurisdiction`,
or the :class:`.Court` asserts that the validity of the :class:`.Rule`
should be considered undecided.
"""

from __future__ import annotations

from itertools import chain
import operator
from typing import Any, Dict, Iterator, List, Tuple
from typing import Optional, Union

from dataclasses import dataclass

from authorityspoke.explanations import Explanation
from authorityspoke.factors import Factor, new_context_helper
from authorityspoke.factors import Analogy, ContextRegister
from authorityspoke.formatting import indented, wrapped
from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule


@dataclass(frozen=True)
class Holding(Factor):
    """
    An :class:`.Opinion`\'s announcement that it posits or rejects a legal :class:`.Rule`.

    Note that if an opinion merely says the court is not deciding whether
    a :class:`.Rule` is valid, there is no :class:`Holding`, and no
    :class:`.Rule` object should be created. Deciding not to decide
    a :class:`Rule`\'s validity is not the same thing as deciding
    that the :class:`.Rule` is undecided.

    :param rule:
        a statement of a legal doctrine about a :class:`.Procedure` for litigation.

    :param rule_valid:
        ``True`` means the :class:`.Rule` is asserted to be valid (or
        useable by a court in litigation). ``False`` means it's asserted
        to be invalid.

    :param decided:
        ``False`` means that it should be deemed undecided
        whether the :class:`.Rule` is valid, and thus can have the
        effect of overruling prior holdings finding the :class:`.Rule`
        to be either valid or invalid. Seemingly, ``decided=False``
        should render the ``rule_valid`` flag irrelevant.

    """

    rule: Rule
    rule_valid: bool = True
    decided: bool = True
    exclusive: bool = False
    generic: bool = False

    def __post_init__(self):
        if self.exclusive:
            if not self.rule_valid:
                raise NotImplementedError(
                    "The ability to state that it is not 'valid' to assert "
                    + "that a Rule is the 'exclusive' way to reach an output is "
                    + "not implemented, so 'rule_valid' cannot be False while "
                    + "'exclusive' is True. Try expressing this in another way "
                    + "without the 'exclusive' keyword."
                )
            if not self.decided:
                raise NotImplementedError(
                    "The ability to state that it is not 'decided' whether "
                    + "a Rule is the 'exclusive' way to reach an output is "
                    + "not implemented. Try expressing this in another way "
                    + "without the 'exclusive' keyword."
                )

    @property
    def procedure(self):
        return self.rule.procedure

    @property
    def despite(self):
        return self.rule.procedure.despite

    @property
    def inputs(self):
        return self.rule.procedure.inputs

    @property
    def outputs(self):
        return self.rule.procedure.outputs

    @property
    def enactments(self):
        return self.rule.enactments

    @property
    def enactments_despite(self):
        return self.rule.enactments_despite

    @property
    def context_factors(self) -> Tuple:
        r"""
        Call :class:`Procedure`\'s :meth:`~Procedure.context_factors` method.

        :returns:
            context_factors from ``self``'s :class:`Procedure`
        """
        return self.rule.procedure.context_factors

    @property
    def generic_factors(self) -> List[Factor]:
        r"""
        Get :class:`.Factor`\s that can be replaced without changing ``self``\s meaning.

        :returns:
            generic :class:`.Factor`\s from ``self``'s :class:`Procedure`
        """
        return self.rule.generic_factors

    @property
    def mandatory(self) -> bool:
        return self.rule.mandatory

    @property
    def universal(self) -> bool:
        return self.rule.universal

    def add_if_not_exclusive(self, other: Holding) -> Optional[Holding]:
        new_rule = self.rule + other.rule
        if new_rule is None:
            return None
        return self.evolve({"rule": self.rule + other.rule})

    def add_holding(self, other: Holding) -> Optional[Holding]:
        if not (self.decided and other.decided):
            raise NotImplementedError(
                "Adding is not implemented for Holdings that assert a Rule is not decided."
            )
        if not (self.rule_valid and other.rule_valid):
            raise NotImplementedError(
                "Adding is not implemented for Holdings that assert a Rule is not valid."
            )
        for self_holding in self.nonexclusive_holdings:
            for other_holding in other.nonexclusive_holdings:
                added = self_holding.add_if_not_exclusive(other_holding)
                if added is not None:
                    return added
        return None

    def __add__(self, other: Factor) -> Optional[Union[Rule, Holding]]:
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            return self.add_holding(other)
        return self.evolve({"rule": self.rule + other})

    def _explanations_contradiction_of_holding(
        self, other: Holding, context: ContextRegister
    ) -> Iterator[Explanation]:
        for self_holding in self.nonexclusive_holdings:
            for other_holding in other.nonexclusive_holdings:
                for register in self_holding._contradicts_if_not_exclusive(
                    other_holding, context=context
                ):
                    yield Explanation(
                        needs_match=self_holding,
                        available=other_holding,
                        context=register,
                        relation="CONTRADICTION",
                    )

    def explanations_contradiction(
        self, other: Factor, context: ContextRegister = None
    ) -> Iterator[Explanation]:
        r"""
        Find context matches that would result in a contradiction with other.

        :param other:
            The :class:`.Factor` to be compared to self. Unlike with
            :meth:`~Holding.contradicts`\, this method cannot be called
            with an :class:`.Opinion` for `other`.

        :returns:
            a generator yielding :class:`.ContextRegister`\s that cause a
            contradiction.
        """

        if context is None:
            context = ContextRegister()
        if isinstance(other, Procedure):
            other = Rule(procedure=other)
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, self.__class__):
            yield from self._explanations_contradiction_of_holding(other, context)
        elif isinstance(other, Factor):
            yield from []  # no possible contradiction
        elif hasattr(other, "explanations_contradiction"):
            if context:
                context = context.reversed()
            yield from other.explanations_contradiction(self, context=context)
        else:
            raise TypeError(
                f"'Contradicts' test not implemented for types "
                f"{self.__class__} and {other.__class__}."
            )

    def _contradicts_if_not_exclusive(
        self, other: Holding, context: ContextRegister = None
    ) -> Iterator[ContextRegister]:
        if context is None:
            context = ContextRegister()
        if isinstance(other, Holding) and other.decided:
            if self.decided:
                yield from self._explanations_implies_if_not_exclusive(
                    other.negated(), context=context
                )
            else:
                yield from chain(
                    other._implies_if_decided(self),
                    other._implies_if_decided(self.negated()),
                )

    def contradicts(
        self, other: Union[Factor, "Opinion"], context: ContextRegister = None
    ) -> bool:
        r"""
        Test if ``self`` :meth:`~.Factor.implies` ``other`` :meth:`~.Factor.negated`\.

        Works by testing whether ``self`` would imply ``other`` if
        ``other`` had an opposite value for ``rule_valid``.

        This method takes three main paths depending on
        whether the holdings ``self`` and ``other`` assert that
        rules are decided or undecided.

        A ``decided`` :class:`Rule` can never contradict
        a previous statement that any :class:`Rule` was undecided.

        If rule A implies rule B, then a holding that B is undecided
        contradicts a prior :class:`Rule` deciding that
        rule A is valid or invalid.

        :returns:
            whether ``self`` contradicts ``other``.
        """
        if context is None:
            context = ContextRegister()
        return any(self.explanations_contradiction(other, context))

    def _explanations_implies_if_not_exclusive(
        self, other: Factor, context: ContextRegister = None
    ) -> Iterator[ContextRegister]:
        if not isinstance(other, self.__class__):
            raise TypeError

        if self.decided and other.decided:
            yield from self._implies_if_decided(other, context)

        # A holding being undecided doesn't seem to imply that
        # any other holding is undecided, except itself and the
        # negation of itself.

        # It doesn't seem that any holding being undecided implies
        # that any holding is decided, or vice versa.

        elif not self.decided and not other.decided:
            yield from chain(
                self.explanations_same_meaning(other, context),
                self.explanations_same_meaning(other.negated(), context),
            )

    def implies(self, other: Factor, context: ContextRegister = None) -> bool:
        r"""
        Test for implication.

        See :meth:`.Procedure.implies_all_to_all`
        and :meth:`.Procedure.implies_all_to_some` for
        explanations of how ``inputs``, ``outputs``,
        and ``despite`` :class:`.Factor`\s affect implication.

        :param other:
            A :class:`Holding` to compare to self, or a :class:`.Rule` to
            convert into such a :class:`Holding` and then compare

        :returns:
            whether ``self`` implies ``other``
        """
        if other is None:
            return True
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if not isinstance(other, self.__class__):
            if hasattr(other, "implied_by"):
                if context:
                    context = context.reversed()
                return other.implied_by(self, context=context)
            return False
        return any(self.explanations_implication(other, context))

    def _context_registers_for_implication(
        self, other: Holding, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        if self.exclusive is other.exclusive is False:
            yield from self._explanations_implies_if_not_exclusive(
                other, context=context
            )
        else:
            analogy = Analogy(
                need_matches=other.nonexclusive_holdings,
                available=self.nonexclusive_holdings,
                comparison=operator.le,
            )
            # in the "exclusive" case, should this return all relevant
            # matches from the nonexclusive Holdings?
            yield from analogy.unordered_comparison(matches=context)

    def explanations_implication(
        self, other: Holding, context: Optional[ContextRegister] = None
    ) -> Iterator[Explanation]:
        for register in self._context_registers_for_implication(other, context=context):
            yield Explanation(self, other, register, relation="IMPLICATION")

    def implied_by(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> bool:
        r"""
        Test if other implies self.

        This function is for handling implication checks for classes
        that don't know the structure of the :class:`Holding` class,
        such as :class:`.Fact` and :class:`.Rule`\.
        """
        if context:
            context = context.reversed()
        if isinstance(other, Rule):
            return Holding(rule=other).implies(self, context=context)
        return other.implies(self, context=context)

    def __ge__(self, other: Union[Holding, Rule]) -> bool:
        return self.implies(other)

    def _implies_if_decided(
        self, other: Holding, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        r"""
        Test if ``self`` implies ``other`` if they're both decided.

        This is a partial version of the
        :meth:`Holding.__ge__` implication function.

        :returns:
            whether ``self`` implies ``other``, assuming that
            ``self.decided == other.decided == True`` and that
            ``self`` and ``other`` are both :class:`Holding`\s,
            although ``rule_valid`` can be ``False``.
        """

        if self.rule_valid and other.rule_valid:
            yield from self.rule.explanations_implication(other.rule, context)

        elif not self.rule_valid and not other.rule_valid:
            yield from other.rule.explanations_implication(self.rule, context)

        # Looking for implication where self.rule_valid != other.rule_valid
        # is equivalent to looking for contradiction.

        # If decided rule A contradicts B, then B also contradicts A

        else:
            yield from self.rule.explanations_contradiction(other.rule, context)

    def __len__(self):
        r"""
        Count generic :class:`.Factor`\s needed as context for this :class:`Holding`.

        :returns:
            the number of generic :class:`.Factor`\s needed for
            self's :class:`.Procedure`.
        """

        return len(self.rule.procedure)

    @property
    def inferred_from_exclusive(self) -> List[Holding]:
        r"""
        Yield :class:`Holding`\s that can be inferred from the "exclusive" flag.

        The generator will be empty if `self.exclusive` is False.
        """
        if self.exclusive:
            return [
                Holding(rule=modified_rule)
                for modified_rule in self.rule.get_contrapositives()
            ]
        return []

    def evolve(self, changes: Union[str, Tuple[str, ...], Dict[str, Any]]) -> Holding:
        """
        Make new object with attributes from ``self.__dict__``, replacing attributes as specified.

        :param changes:
            a :class:`dict` where the keys are names of attributes
            of self, and the values are new values for those attributes, or
            else an attribute name or :class:`list` of names that need to
            have their values replaced with their boolean opposite.

        :returns:
            a new object initialized with attributes from
            ``self.__dict__``, except that any attributes named as keys in the
            changes parameter are replaced by the corresponding value.
        """
        changes = self._make_dict_to_evolve(changes)
        changes = self._evolve_attribute(changes, "rule")
        new_values = self._evolve_from_dict(changes)
        return self.__class__(**new_values)

    def explanations_same_meaning(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        if (
            isinstance(other, self.__class__)
            and self.rule_valid == other.rule_valid
            and self.decided == other.decided
        ):
            yield from self.rule.explanations_same_meaning(other.rule, context)

    def means(
        self, other: Optional[Factor], context: Optional[ContextRegister] = None
    ) -> bool:
        """
        Test whether ``other`` has the same meaning as ``self``.

        :returns:
            whether ``other`` is a :class:`Holding` with the
            same meaning as ``self``.
        """
        if other is None:
            return False
        return any(self.explanations_same_meaning(other, context))

    def negated(self):
        """Get new copy of ``self`` with an opposite value for ``rule_valid``."""
        return self.evolve({"rule_valid": not self.rule_valid, "exclusive": False})

    @new_context_helper
    def new_context(self, changes: ContextRegister) -> Factor:
        """
        Create new :class:`Holding`, replacing keys of ``changes`` with values.

        :returns:
            a version of ``self`` with the new context.
        """
        return Holding(
            rule=self.rule.new_context(changes),
            rule_valid=self.rule_valid,
            decided=self.decided,
        )

    @property
    def nonexclusive_holdings(self) -> List[Holding]:
        r"""
        Yield all :class:`.Holding`\s with `exclusive is False` implied by self.
        """
        if not self.exclusive:
            return [self]
        return [self.evolve("exclusive")] + self.inferred_from_exclusive

    def _union_if_not_exclusive(
        self, other: Holding, context: ContextRegister
    ) -> Optional[Holding]:
        if self.decided is other.decided is False:
            if self.rule.implies(other.rule, context=context):
                return other
            if other.rule.implies(self.rule, context=context.reversed()):
                return self
            return None

        if not self.decided or not other.decided:
            return None
        if self.rule_valid != other.rule_valid:
            return None

        if self.rule_valid is False:
            # If a Rule with input A present is not valid
            # and a Rule with input A absent is also not valid
            # then a version of the Rule with input A
            # omitted is also not valid.
            raise NotImplementedError(
                "The union operation is not yet implemented for Holdings "
                "that assert a Rule is not valid."
            )

        new_rule = self.rule.union(other.rule, context=context)
        if not new_rule:
            return None
        return self.evolve({"rule": new_rule, "exclusive": False})

    def _union_with_holding(
        self, other: Holding, context: ContextRegister
    ) -> Optional[Holding]:
        for self_holding in self.nonexclusive_holdings:
            for other_holding in other.nonexclusive_holdings:
                united = self_holding._union_if_not_exclusive(
                    other_holding, context=context
                )
                if united is not None:
                    return united
        return None

    def union(
        self, other: Union[Rule, Holding], context: Optional[ContextRegister] = None
    ) -> Optional[Holding]:
        context = context or ContextRegister()
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if not isinstance(other, Holding):
            raise TypeError
        return self._union_with_holding(other, context=context)

    def __or__(self, other: Union[Rule, Holding]) -> Optional[Holding]:
        return self.union(other)

    def own_attributes(self) -> Dict[str, Any]:
        """
        Return attributes of ``self`` that aren't inherited from another class.

        Used for getting parameters to pass to :meth:`~Holding.__init__`
        when generating a new object.
        """
        attrs = self.__dict__.copy()
        for group in Procedure.context_factor_names:
            attrs.pop(group, None)
        for group in Rule.enactment_attr_names:
            attrs.pop(group, None)
        attrs.pop("procedure", None)
        return attrs

    def __str__(self):
        action = (
            "consider UNDECIDED"
            if not self.decided
            else ("ACCEPT" if self.rule_valid else "REJECT")
        )
        exclusive = (
            (
                f" that the EXCLUSIVE way to reach "
                f"{self.rule.outputs[0].short_string} is"
            )
            if self.exclusive
            else ""
        )
        rule_text = indented(str(self.rule))
        text = wrapped(f"the Holding to {action}{exclusive}") + f"\n{rule_text}"
        return text



import pytest

from authorityspoke.entities import Entity
from authorityspoke.io import loaders, readers
from authorityspoke.opinions import Opinion


class TestOpinions:
    def test_opinion_features(self, make_decision):
        assert make_decision["watt"].court == "9th-cir"
        assert make_decision["watt"].citations[0].cite == "388 F.2d 853"

    def test_repr(self, make_opinion):
        assert "HAMLEY, Circuit Judge" in repr(make_opinion["watt_majority"])

    def test_repr_excludes_text(self, make_opinion):
        """
        reprs would be way too long if they could contain the
        full opinion text.
        """
        assert "text" not in repr(make_opinion["watt_majority"])

    def test_opinion_author(self, make_opinion):
        assert make_opinion["watt_majority"].author == "HAMLEY, Circuit Judge"
        assert make_opinion["brad_majority"].author == "BURKE, J."
        assert (
            make_opinion["brad_concurring-in-part-and-dissenting-in-part"].author
            == "TOBRINER, J."
        )

    def test_opinion_holding_list(self, make_opinion, real_holding):
        watt = make_opinion["watt_majority"]
        h3_specific = real_holding["h3"]
        watt.posit(h3_specific)
        assert h3_specific in watt.holdings

    def test_opinion_entity_list(
        self, make_opinion, real_holding, make_entity, make_evidence
    ):
        watt = make_opinion["watt_majority"]
        h = real_holding
        e = make_entity

        watt.posit(h["h1"], context=(e["motel"], e["watt"]))
        watt.posit(h["h2"], context=(e["trees"], e["motel"]))
        watt.posit(
            h["h3"],
            context=(
                make_evidence["generic"],
                e["motel"],
                e["watt"],
                e["trees"],
                e["tree_search"],
            ),
        )
        watt.posit(
            h["h4"], context=(e["trees"], e["tree_search"], e["motel"], e["watt"])
        )
        assert make_entity["watt"] in make_opinion["watt_majority"].generic_factors

    def test_opinion_date(self, make_decision):
        """
        This no longer tests whether two Opinions from the same Decision
        have the same date, because the date field has been moved to the
        Decision class.
        """
        assert make_decision["watt"].date < make_decision["brad"].date

    def test_opinion_string(self, make_opinion):
        opinion = make_opinion["cardenas_majority"]
        assert str(opinion).lower() == "majority opinion by bird, c. j."


class TestOpinionText:
    def test_opinion_text(self, make_opinion):
        assert (
            "Feist responded that such efforts were economically "
            + "impractical and, in any event, unnecessary"
        ) in make_opinion["feist_majority"].text

    def test_opinion_text_anchor(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        assert any(
            "generally" in anchor.exact
            for anchor in feist.get_anchors(feist.holdings[1])
        )

    def test_opinion_factor_text_anchor(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        anchors = feist.get_anchors(feist.holdings[0])
        assert all(
            "No one may claim originality" not in anchor.exact for anchor in anchors
        )
        assert any("as to facts" in anchor.exact for anchor in anchors)

    def test_select_opinion_text_for_factor(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        factor = oracle.holdings[0].outputs[0]
        anchor = oracle.factors[factor][0]
        selected = oracle.select_text(selector=anchor)
        assert selected == "copyright protection."

    def test_select_opinion_text_for_enactment(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        enactment = oracle.holdings[0].enactments[0]
        anchor = oracle.factors[enactment][0]
        selected = oracle.select_text(selector=anchor)
        assert selected == "17 U.S.C. § 102(a)"

    def test_select_opinion_text_for_holding(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        holding = oracle.holdings[0]
        anchor = oracle.holding_anchors[holding][0]
        selected = oracle.select_text(selector=anchor)
        assert selected == "must be “original” to qualify"


class TestOpinionHoldings:
    def test_positing_non_rule_error(self, make_opinion, make_procedure):
        with pytest.raises(TypeError):
            make_opinion["watt_majority"].posit(make_procedure["c1"])

    def test_error_posit_with_no_rule_source(self, make_opinion):
        with pytest.raises(TypeError):
            make_opinion["watt_majority"].posit()

    def test_posit_rule(self, make_opinion, make_rule, make_holding):
        """
        "Positing" a Rule causes the Rule to be converted to a Holding first.
        So the Opinion implies the corresponding Holding.
        """

        watt = make_opinion["watt_majority"]
        watt.posit(make_rule["h1"])
        assert watt.implies(make_holding["h1"])

    def test_new_context_non_iterable_changes(self, make_opinion, make_holding):
        """
        The context here (a Factor outside an iterable) only changes the first
        generic factor of the Rule being posited, which may not be what the user
        expects.
        """
        brad = make_opinion["brad_majority"]
        brad.posit(make_holding["h1"], context=Entity("House on Haunted Hill"))
        assert "Haunted Hill" in str(brad.holdings[0])

    def test_new_context_naming_nonexistent_factor(self, make_opinion, make_holding):
        """
        The context here (a Factor outside an iterable) only changes the first
        generic factor of the Rule being posited, which may not be what the user
        expects.
        """
        brad = make_opinion["brad_majority"]
        with pytest.raises(ValueError):
            brad.posit(
                make_holding["h1"],
                context=(Entity("House on Haunted Hill"), "nonexistent factor"),
            )

    def test_new_context_creates_equal_rule(self, make_opinion, make_regime):
        watt = make_opinion["watt_majority"]
        brad = make_opinion["brad_majority"]

        watt.clear_holdings()
        watt_raw = loaders.load_holdings("holding_watt.json")
        watt.posit(readers.read_holdings(watt_raw, regime=make_regime))

        brad.clear_holdings()
        brad_raw = loaders.load_holdings("holding_brad.json")
        brad.posit(readers.read_holdings(brad_raw, regime=make_regime))

        context_pairs = {
            "proof of Bradley's guilt": "proof of Wattenburg's guilt",
            "Bradley": "Wattenburg",
            "officers' search of the yard": "officers' search of the stockpile",
            "Bradley's marijuana patch": "the stockpile of trees",
        }
        watt.posit(brad.holdings[0], context_pairs)
        assert watt.holdings[-1].means(brad.holdings[0])

    def test_new_context_inferring_factors_to_change(self, make_opinion, make_regime):
        """
        This changes watt's holdings; may break tests below.
        """

        watt = make_opinion["watt_majority"]
        brad = make_opinion["brad_majority"]

        watt.clear_holdings()
        watt_raw = loaders.load_holdings("holding_watt.json")
        watt.posit(readers.read_holdings(watt_raw, regime=make_regime))

        brad.clear_holdings()
        brad_raw = loaders.load_holdings("holding_brad.json")
        brad.posit(readers.read_holdings(brad_raw, regime=make_regime))

        context_items = [
            "proof of Wattenburg's guilt",
            "Wattenburg",
            "officers' search of the stockpile",
            "Hideaway Lodge",
            "the stockpile of trees",
        ]
        watt.posit(brad.holdings[0], context=context_items)
        assert watt.holdings[-1].means(brad.holdings[0])


class TestImplication:
    def test_no_implication(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        brad = make_opinion_with_holding["brad_majority"]
        assert not watt >= brad

    def test_posit_list_of_holdings_and_imply(self, make_opinion, make_regime):
        watt = make_opinion["watt_majority"]
        brad = make_opinion["brad_majority"]
        some_rules_raw = loaders.load_holdings(filename="holding_watt.json")
        some_rules = readers.read_holdings(some_rules_raw, regime=make_regime)
        for case in (watt, brad):
            case.clear_holdings()
            case.posit(some_rules[:3])
        watt.posit(some_rules[3])
        assert watt > brad
        assert not brad >= watt

    def test_opinion_implies_holding(self, make_opinion, make_holding):
        watt = make_opinion["watt_majority"]
        watt.posit(make_holding["h2_invalid_undecided"])
        assert watt >= make_holding["h2_undecided"]
        assert watt > make_holding["h2_undecided"]

    def test_opinion_does_not_imply_holding(self, make_opinion, make_holding):
        watt = make_opinion["watt_majority"]
        watt.clear_holdings()
        watt.posit(make_holding["h2_irrelevant_inputs_undecided"])
        assert not watt >= make_holding["h2_undecided"]
        assert not watt > make_holding["h2_undecided"]


class TestContradiction:
    def test_contradiction_of_holding(
        self, make_opinion_with_holding, make_enactment, make_holding
    ):
        assert make_opinion_with_holding["watt_majority"].contradicts(
            make_holding["h2_output_false_ALL_MUST"] + make_enactment["search_clause"]
        )

    def test_explain_opinion_contradicting_holding(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        explanation = oracle.explain_contradiction(lotus.holdings[6])
        assert "an explanation" in str(explanation).lower()

    def test_contradiction_of_decision(
        self, make_opinion_with_holding, make_decision_with_holding
    ):
        assert make_opinion_with_holding["oracle_majority"].contradicts(
            make_decision_with_holding["lotus"]
        )

    def test_explain_opinion_contradicting_decision(
        self, make_opinion_with_holding, make_decision_with_holding
    ):
        oracle_majority = make_opinion_with_holding["oracle_majority"]
        lotus = make_decision_with_holding["lotus"]
        explanation = oracle_majority.explain_contradiction(lotus)
        assert "an explanation" in str(explanation).lower()

    def test_error_contradiction_with_procedure(self, make_opinion, make_procedure):
        with pytest.raises(TypeError):
            make_opinion["watt_majority"].contradicts(make_procedure["c1"])



"""AuthoritySpoke: Reading the law for the last time."""

from .enactments import Code, Enactment
from .entities import Entity
from .factors import Factor
from .jurisdictions import Jurisdiction, Regime
from .opinions import Opinion
from .rules import Rule
from .io.dump import to_dict, to_json

__version__ = "0.3.0"



"""Functions for indexing named objects in JSON to be imported."""
from __future__ import annotations

from collections import OrderedDict

from typing import Dict, List, Optional, Union

from marshmallow import ValidationError


class Mentioned(OrderedDict):
    def insert_by_name(self, obj: Dict) -> None:
        self[obj["name"]] = obj.copy()
        self[obj["name"]].pop("name")
        return None

    def get_by_name(self, name: str) -> Dict:
        if not self.get(name):
            raise ValidationError(
                f'Name "{name}" not found in the index of mentioned Factors'
            )
        value = {"name": name}
        value.update(self[name])
        return value

    def sorted_by_length(self) -> Mentioned:
        return Mentioned(sorted(self.items(), key=lambda t: len(t[0]), reverse=True))

    def __str__(self):
        return f"Mentioned({str(dict(self))})"

    def __repr__(self):
        return f"Mentioned({repr(dict(self))})"


def collect_mentioned(obj: Dict, mentioned: Optional[Mentioned] = None) -> Mentioned:
    """
    Make a dict of all nested objects labeled by name.

    To be used during loading to expand name references to full objects.
    """
    mentioned = mentioned or Mentioned()
    if isinstance(obj, List):
        for item in obj:
            mentioned.update(collect_mentioned(item))
    if isinstance(obj, Dict):
        if obj.get("name"):
            mentioned.insert_by_name(obj)
        for _, value in obj.items():
            if isinstance(value, (Dict, List)):
                mentioned.update(collect_mentioned(value))
    return mentioned


def index_names(obj: Union[List, Dict]) -> Mentioned:
    """
    Call all functions to prepare "mentioned" index.

    The names are sorted by length so that if one mentioned Factor's name
    is a substring of another, the longest available name is expanded.

    :returns:
        a modified version of the dict to load, plus a dict of names
        and the objects to expand them with.
    """
    mentioned = collect_mentioned(obj)
    sorted_mentioned = mentioned.sorted_by_length()
    return sorted_mentioned



import textwrap


TAB_WIDTH = 2
TEXT_WIDTH = 70


def indented(text: str, tabs: int = 1) -> str:
    indent = TAB_WIDTH * tabs * " "
    return textwrap.indent(text, prefix=indent)


def wrapped(text: str) -> str:
    return textwrap.fill(text, width=TEXT_WIDTH)



r"""
:class:`Court` documents that decide litigation and posit :class:`.Rule`\s.

Unlike most other ``authorityspoke`` classes, :class:`Opinion`\s are not frozen.
"""

from __future__ import annotations

from collections import defaultdict
from itertools import zip_longest
import operator
from typing import Dict, Iterable, Iterator, List
from typing import Optional, Sequence, Set, Tuple, Union

import logging
import re

from dataclasses import dataclass, field

from anchorpoint.textselectors import TextQuoteSelector

from authorityspoke.factors import Factor, ContextRegister, Analogy
from authorityspoke.explanations import Explanation
from authorityspoke.holdings import Holding
from authorityspoke.rules import Rule


logger = logging.getLogger(__name__)


TextLinkDict = Dict[str, List[TextQuoteSelector]]


@dataclass
class Opinion:
    """
    A document that resolves legal issues in a case and posits legal holdings.

    Usually an opinion must have ``position="majority"``
    to create holdings binding on any courts.

    :param position:
        the opinion's attitude toward the court's disposition of the case.
        e.g. ``majority``, ``dissenting``, ``concurring``, ``concurring in the result``
    :param author:
        name of the judge who authored the opinion, if identified
    :param text:
    """

    position: Optional[str] = None
    author: Optional[str] = None
    text: Optional[str] = field(default=None, repr=False)

    def __post_init__(self):
        r"""
        Add attributes to store Holdings and Factors keyed to their text selectors.

        The ``holding_anchors`` are text selectors for the whole :class:`Holding`, not for any
        individual :class:`.Factor`. Often select text used to
        indicate whether the :class:`.Rule` is ``mandatory``, ``universal``,
        ``valid``, or ``decided``, or show the ``exclusive`` way to reach
        the outputs.
        """

        self.holding_anchors: Dict[Holding, List[TextQuoteSelector]] = defaultdict(list)
        self.factors: Dict[Factor, List[TextQuoteSelector]] = defaultdict(list)

    def clear_holdings(self):
        r"""
        Remove all :class:`.Holding`\s from the opinion.
        """
        self.holding_anchors.clear()

    def explain_contradiction(
        self,
        other: Union[Opinion, Holding, Rule],
        context: Optional[ContextRegister] = None,
    ) -> Optional[Explanation]:
        explanations = self.explanations_contradiction(other, context=context)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explanations_contradiction(
        self,
        other: Union[Opinion, Holding, Rule],
        context: Optional[ContextRegister] = None,
    ) -> Iterator[Explanation]:
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            for self_holding in self.holdings:
                for explanation in self_holding.explanations_contradiction(
                    other, context
                ):
                    yield explanation
        elif isinstance(other, self.__class__):
            for self_holding in self.holdings:
                for other_holding in other.holdings:
                    for explanation in self_holding.explanations_contradiction(
                        other_holding, context
                    ):
                        yield explanation
        elif hasattr(other, "explanations_contradiction"):
            if context:
                context = context.reversed()
            yield from other.explanations_contradiction(self, context=context)
        else:
            raise TypeError(
                f"'Contradicts' test not implemented for types "
                + f"{self.__class__} and {other.__class__}."
            )

    def explain_implication(
        self,
        other: Union[Opinion, Holding, Rule],
        context: Optional[ContextRegister] = None,
    ) -> Optional[Explanation]:
        explanations = self.explanations_implication(other, context=context)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explanations_implication(
        self,
        other: Union[Opinion, Holding, Rule],
        context: Optional[ContextRegister] = None,
    ) -> Iterator[Explanation]:
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            for self_holding in self.holdings:
                for explanation in self_holding.explanations_implication(
                    other, context
                ):
                    yield explanation
        elif isinstance(other, self.__class__):
            analogy = Analogy(
                need_matches=other.holdings,
                available=self.holdings,
                comparison=operator.le,
            )
            yield from analogy.unordered_comparison(matches=context)
        elif hasattr(other, "explanations_implication"):
            if context:
                context = context.reversed()
            yield from other.explanations_implication(self, context=context)
        else:
            raise TypeError(
                f"'Implies' test not implemented for types {self.__class__} and {other.__class__}."
            )

    def contradicts(
        self,
        other: Union[Opinion, Holding, Rule],
        context: Optional[ContextRegister] = None,
    ) -> bool:
        """
        Test whether ``other`` is or contains a :class:`.Holding` contradicted by ``self``.

        :param other:
            another :class:`.Opinion` or :class:`.Holding`

        :returns:
            a bool indicating whether any holding of ``self`` is
            inconsistent with the :class:`.Rule` ``other``, or with
            any holding of ``other`` if ``other`` is an :class:`.Opinion`.
        """

        return any(self.explanations_contradiction(other, context=context))

    @property
    def generic_factors(self) -> List[Factor]:
        r"""
        Get all generic :class:`.Factor`\s mentioned in ``self``.

        :returns:
            a list of generic :class:`.Factor` objects mentioned in
            any ``input``, ``output``, or ``despite`` :class:`.Factor`
            of ``self``, with guaranteed order, including each
            generic :class:`.Factor` only once.
        """
        return list(
            {
                generic: None
                for holding in self.holdings
                for generic in holding.generic_factors
            }
        )

    def get_factor_by_name(self, name: str) -> Optional[Factor]:
        """
        Search recursively in holdings of ``self`` for :class:`.Factor` with ``name``.

        :param name:
            string to match with the ``name`` attribute of the
            :class:`.Factor` being searched for.

        :returns:
            a :class:`.Factor` with the specified ``name``, if one
            exists in a :class:`.Rule` in ``self.holdings``.
            Otherwise ``None``.
        """

        for holding in self.holdings:
            factor = holding.get_factor_by_name(name)
            if factor is not None:
                return factor
        return None

    @property
    def holdings(self) -> List[Holding]:
        r"""
        Get ordered list of :class:`.Holding`\s posited by this :class:`Opinion`

        :returns:
            keys of the holding_anchors :class:`.OrderedDict`, as a list
        """
        return list(self.holding_anchors)

    def posit_holding(
        self,
        holding: Union[Holding, Rule],
        holding_anchors: Optional[List[TextQuoteSelector]] = None,
        named_anchors: Optional[TextLinkDict] = None,
        context: Optional[Sequence[Factor]] = None,
    ) -> None:
        r"""Record that this Opinion endorses specified :class:`Holding`\s."""
        if isinstance(holding, Rule):
            logger.warning(
                "posit_holding was called with a Rule "
                "that was automatically converted to a Holding"
            )
            holding = Holding(rule=holding)

        if not isinstance(holding, Holding):
            raise TypeError('"holding" must be an object of type Holding.')

        if context:
            holding = holding.new_context(context, context_opinion=self)
        self.holding_anchors[holding].extend(holding_anchors or [])
        if named_anchors:
            for factor in (
                list(holding.recursive_factors)
                + list(holding.enactments)
                + list(holding.enactments_despite)
            ):
                if hasattr(factor, "name") and factor.name in named_anchors:
                    self.factors[factor].extend(named_anchors[factor.name])

    def posit_holdings(
        self,
        holdings: Iterable[Union[Holding, Rule]],
        holding_anchors: Optional[List[List[TextQuoteSelector]]] = None,
        named_anchors: Optional[TextLinkDict] = None,
        context: Optional[Sequence[Factor]] = None,
    ):
        r"""
        Add :class:`.Holding`\s to this ``Opinion`` from a sequence.

        :param holdings:
            a sequence of :class:`.Holding`\s that this :class:`.Opinion`
            posits as valid in its own court or jurisdiction, regardless of
            whether ``self`` accepts that the ``inputs`` correspond to the
            reality of the current case, and regardless of whether the
            court orders that the ``outputs`` be put into effect.

        :param text_links:
            mapping of :class:`Factor`\s to the :class:`Opinion` passages where
            they can be found. Can be obtained as the `mentioned` return value
            of one of the functions in :mod:`authorityspoke.io.readers`\.

        :param context:
            an ordered sequence (probably :py:class:`dict`) of
            generic :class:`.Factor` objects from ``self`` which
            will provide the context for the new holding in the
            present case.
        """
        holding_anchors = holding_anchors or []
        for (holding, selector_list) in zip_longest(holdings, holding_anchors):
            self.posit_holding(
                holding,
                holding_anchors=selector_list,
                named_anchors=named_anchors,
                context=context,
            )

    def posit(
        self,
        holdings: Union[Holding, Iterable[Union[Holding, Rule]]],
        holding_anchors: Optional[
            List[Union[TextQuoteSelector, List[TextQuoteSelector]]]
        ] = None,
        named_anchors: Optional[TextLinkDict] = None,
        context: Optional[Sequence[Factor]] = None,
    ) -> None:
        r"""
        Add one or more :class:`.Holding`\s to this ``Opinion``.

        This method passes its values to :meth:`~posit_holding` or
        :meth:`~posit_holdings` depending on whether the `holding` parameter
        is one :class:`.Holding` or a :class:`list`\.

        :param holding:
            a :class:`.Holding` that the :class:`.Opinion` ``self`` posits
            as valid in its own court or jurisdiction, regardless of
            whether ``self`` accepts that the ``inputs`` of the
            :class:`.Holding` correspond to the reality of the current
            case, and regardless of whether the court orders that
            the ``outputs`` of the :class:`.Holding` be put into effect.

        :param text_links:
            list of lists of :class:`.Opinion` passages where references to each
            :class:`.Holding` can be found.

        :param context:
            an ordered sequence (probably :py:class:`dict`) of
            generic :class:`.Factor` objects from ``self`` which
            will provide the context for the new holding in the
            present case.
        """
        if isinstance(holdings, Iterable):
            self.posit_holdings(
                holdings,
                holding_anchors=holding_anchors,
                named_anchors=named_anchors,
                context=context,
            )
        else:
            self.posit_holding(
                holdings,
                holding_anchors=holding_anchors,
                named_anchors=named_anchors,
                context=context,
            )

    def get_anchors(self, holding: Holding, include_factors: bool = True) -> List[str]:
        r"""
        Get text passages where a :class:`.Holding` is linked to ``self``.

        :param holding:
            a holding to find anchors for, which must be in :attr:`~Opinion.holdings`\.

        :returns:
            a :class:`list` with the text of each passage that anchors the :class:`.Holding`
        """

        anchors: List[str] = []
        if include_factors:
            for factor in holding.rule.procedure.factors_all:
                if self.factors.get(factor):
                    anchors.extend(self.factors[factor])
        anchors.extend(self.holding_anchors[holding])

        return anchors

    def implied_by_holding(
        self, other: Holding, context: ContextRegister = None
    ) -> bool:
        return all(
            other.implies(self_holding, context=context.reversed())
            for self_holding in self.holdings
        )

    def implied_by_rule(self, other: Rule, context: ContextRegister = None) -> bool:
        return self.implied_by_holding(other=Holding(other), context=context)

    def implied_by(
        self, other: Union[Opinion, Holding, Rule], context: ContextRegister = None
    ) -> bool:
        if isinstance(other, Holding):
            return self.implied_by_holding(other, context=context)
        elif isinstance(other, Rule):
            return self.implied_by_rule(other, context=context)
        return other.implies(self, context=context.reversed())

    def implies_other_holdings(
        self, other_holdings: List[Holding], context: ContextRegister = None
    ):
        for other_holding in other_holdings:
            if not any(
                self_holding.implies(other_holding, context=context)
                for self_holding in self.holdings
            ):
                return False
        return True

    def implies(
        self, other: Union[Opinion, Holding, Rule], context: ContextRegister = None
    ) -> bool:
        if isinstance(other, (Rule, Holding)):
            return any(
                self_holding.implies(other, context=context)
                for self_holding in self.holdings
            )
        elif isinstance(other, self.__class__):
            return self.implies_other_holdings(other.holdings)
        if hasattr(other, "implied_by"):
            if context:
                context = context.reversed()
            return other.implied_by(self, context=context)
        return False

    def __ge__(self, other: Union[Opinion, Rule]) -> bool:
        """
        Find whether ``self``'s holdings imply all the holdings of ``other``.

        :returns:
            a bool indicating whether the :class:`.Rule` ``other``
            (or every holding of ``other``, if other is an :class:`.Opinion`)
            is implied by some :class:`.Rule` in ``self.holdings``.
        """
        return self.implies(other)

    def __gt__(self, other) -> bool:
        """
        Find whether ``self``\'s holdings imply ``other``\'s but self != other.

        This actually tests for inequality because ``Opinion`` does not
        have a ``means`` method.

        :returns:
            self >= other and self != other.
        """
        return (self >= other) and (self != other)

    def select_text(self, selector: TextQuoteSelector) -> Optional[str]:
        r"""
        Get text using a :class:`.TextQuoteSelector`.

        :param selector:
            a selector referencing a text passage in this :class:`Opinion`.

        :returns:
            the text referenced by the selector, or ``None`` if the text
            can't be found.
        """
        if re.search(selector.passage_regex(), self.text, re.IGNORECASE):
            return selector.exact
        raise ValueError(
            f'Passage "{selector.exact}" from TextQuoteSelector '
            + f'not found in Opinion "{self}".'
        )

    def __str__(self):
        return f"{self.position} Opinion by {self.author}"



r"""
Representations of legal systems.

These clasess are currently used to organize legislation,
but later they can be linked to :class:`.Court`\s and
thus to :class:`.Opinion`\s.
"""

from __future__ import annotations

from typing import ClassVar, Dict, Optional, Union

from dataclasses import dataclass, field

from anchorpoint.textselectors import TextQuoteSelector

from authorityspoke.enactments import Code
from authorityspoke.courts import Court


@dataclass
class Jurisdiction:
    r"""
    A geopolitical entity with the power to enact legislation.

    Has :class:`.Court`\s to interpret and apply legislation.

    :param id:
        the :class:`Jurisdiction`\'s identifier in USLM or, failing that,
        in Citation Style Language.

    :param codes:
        a list of :class:`.Code`\s enacted by the :class:`Jurisdiction`\.

    :param courts:
        a list of :class:`.Court`\s operated by the :class:`Jurisdiction`\.
    """

    codes: Dict[str, Code] = field(default_factory=dict)
    courts: Dict[str, Court] = field(default_factory=dict)

    def get_code(self, uri: str) -> Optional[Code]:
        """
        Find a :class:`.Code` under a key corresponding to ``uri``.

        If the initial form of ``uri`` isn't found, truncates
        parts from the right side of the ``uri`` until a match
        is found or nothing is left.

        :param uri:
            identifier for legislative text in the USLM format
            (or possibly pseudo-USLM for non-USC sources)

        :returns:
            the :class:`.Code` for the ``uri``
        """
        query = uri[:]
        while query:
            if self.codes.get(str(query)):
                return self.codes.get(query)
            query = query[: query.rfind("/")]
        return None


@dataclass
class Regime:
    r"""
    A legal system consisting of multiple jurisdictions.

    Currently used for retrieving :class:`.Enactment` text.
    May be modified for retrieving :class:`.Court`\s and
    :class:`.Opinion`\s.
    """

    jurisdictions: Dict[str, Jurisdiction] = field(default_factory=dict)
    uri: ClassVar = "/"

    def get_code(self, path: str):
        """
        Find a :class:`.Code` in the :class:`Regime` from a selector.

        :returns:
            the :class:`.Code` described in the selector path.
        """
        jurisdiction_id = path.split("/")[1]

        if jurisdiction_id not in self.jurisdictions:
            raise ValueError(f"Regime has no jurisdiction {jurisdiction_id}.")
        return self.jurisdictions[jurisdiction_id].get_code(path)

    def set_code(self, code: Code) -> None:
        r"""
        Add to the collection of :class:`.Code`\s enacted in this :class:`Regime`.

        Create the appropriate :class:`.Jurisdiction` object for the
        :class:`.Code`, if it's not already linked to the :class:`Regime`.

        :param code:
            a :class:`.Code` valid in this :class:`Regime`
        """
        if code.jurisdiction not in self.jurisdictions:
            self.jurisdictions[code.jurisdiction] = Jurisdiction()

        if not self.jurisdictions[code.jurisdiction].codes.get(code.uri):
            self.jurisdictions[code.jurisdiction].codes[code.uri] = code
        return None



r"""
Descriptions of procedural changes in litigation.

Does not specify whether they are mandatory or universal,
or specify the :class:`.Enactment`\s that might require them.
"""

from __future__ import annotations

import functools
import operator

from typing import Callable, ClassVar, Iterator, List, Optional, Tuple

from dataclasses import dataclass

from authorityspoke.factors import Factor, ContextRegister, new_context_helper
from authorityspoke.factors import Analogy, all_analogy_matches, means
from authorityspoke.formatting import indented, wrapped


def find_less_specific_context(
    left: Procedure, right: Procedure, context: Optional[ContextRegister] = None
) -> Optional[ContextRegister]:
    r"""
    Find context assuming :class:`.Factor`\s with same meaning have corresponding generics.

    :param left:
        a :class:`.Procedure` that is being compared to another
        :class:`.Procedure`\, to create a new :class:`.Procedure`
        using the context of the left :class:`.Procedure`\.

    :param right:
        a :class:`.Procedure` that is being compared to another :class:`.Procedure`\,
        but that will have its context overwritten in the newly-created object.

    :param context:
        a :class:`.ContextRegister` identifying any known pairs of corresponding
        context :class:`.Factor`\s between the two :class:`.Procedure`\s being
        compared.

    :returns:
        a :class:`.ContextRegister`\s assuming that :class:`.Factor`\s with the
        same meaning have corresponding generics.
    """
    new_context = context or ContextRegister()
    for left_factor in left.factors_all:
        for right_factor in right.factors_all:
            if left_factor.means(right_factor, context=context):
                incoming = ContextRegister(
                    dict(zip(right_factor.generic_factors, left_factor.generic_factors))
                )
                updated_context = new_context.merged_with(incoming)
                if updated_context is not None:
                    new_context = updated_context
    if new_context and new_context != context:
        return new_context
    return None


def find_more_specific_context(
    left: Procedure, right: Procedure, context: Optional[ContextRegister] = None
) -> Optional[ContextRegister]:
    r"""
    Find context assuming :class:`.Factor`\s that imply each other have corresponding generics.

    :param left:
        a :class:`.Procedure` that is being compared to another
        :class:`.Procedure`\, to create a new :class:`.Procedure`
        using the context of the left :class:`.Procedure`\.

    :param right:
        a :class:`.Procedure` that is being compared to another :class:`.Procedure`\,
        but that will have its context overwritten in the newly-created object.

    :param context:
        a :class:`.ContextRegister` identifying any known pairs of corresponding
        context :class:`.Factor`\s between the two :class:`.Procedure`\s being
        compared.

    :returns:
        a :class:`.ContextRegister`\s assuming that :class:`.Factor`\s with the
        same meaning have corresponding generics.
    """
    new_context = context or ContextRegister()
    for left_factor in left.factors_all:
        for right_factor in right.factors_all:
            if left_factor.implies(
                right_factor, context=context
            ) or right_factor.implies(left_factor, context=context):
                incoming = ContextRegister(
                    dict(zip(right_factor.generic_factors, left_factor.generic_factors))
                )
                updated_context = new_context.merged_with(incoming)
                if updated_context is not None:
                    new_context = updated_context
    if new_context and new_context != context:
        return new_context
    return None


def use_likely_context(func: Callable):
    r"""
    Find contexts most likely to have been intended for comparing :class:`Procedure`\s.

    When such contexts are found, first tries calling the decorated comparison method
    using those contexts. Only if no answer is found using the likely contexts
    will the decorated method be called with no comparison method specified.

    :param left:
        a :class:`.Procedure` that is being compared to another
        :class:`.Procedure`\, to create a new :class:`.Procedure`
        using the context of the left :class:`.Procedure`\.

    :param right:
        a :class:`.Procedure` that is being compared to another :class:`.Procedure`\,
        but that will have its context overwritten in the newly-created object.

    :param context:
        a :class:`.ContextRegister` identifying any known pairs of corresponding
        context :class:`.Factor`\s between the two :class:`.Procedure`\s being
        compared.

    :returns:
        a generator yielding :class:`.ContextRegister`\s based on the most "likely"
        context that yields any :class:`.ContextRegister`\s at all.
    """

    @functools.wraps(func)
    def wrapper(
        left: Procedure, right: Procedure, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        less_specific = find_less_specific_context(left, right, context)
        more_specific = find_more_specific_context(left, right, less_specific)
        context_to_use = more_specific or less_specific or context or ContextRegister()
        for unused_left, unused_right in zip(
            [
                item
                for item in left.generic_factors
                if item not in context_to_use.keys()
            ],
            [
                item
                for item in right.generic_factors
                if item not in context_to_use.values()
            ],
        ):
            context_to_use[unused_right] = unused_left
        return func(left, right, context_to_use)

    return wrapper


@dataclass(frozen=True)
class Procedure(Factor):
    r"""
    A (potential) rule for courts to use in resolving litigation.

    Described in terms of inputs and outputs, and also potentially
    "despite" :class:`.Factor`\s, which occur when a :class:`Rule`
    could be said to apply "even if" some "despite" factor is true.

    Users generally should not need to interact with this class
    directly, under the current design. Instead, they should interact
    with the class :class:`.Rule`.

    :param outputs:
        an outcome that a court may accept based on the presence
        of the ``inputs``

    :param inputs:
        supporting :class:`.Factor`\s in favor of the ``output``.
        The ``input`` :class:`.Factor`\s are not treated as
        potential undercutters.

    :param despite:
        :class:`.Factor`\s that do not prevent the court from
        imposing the ``output``. These could be considered
        "failed undercutters" in defeasible logic. If a :class:`.Factor`
        is relevant both as support for the output and as
        a potential undercutter, include it in both ``inputs``
        and ``despite``.

    :param name:
        An identifier that can be used to reference and
        incorporate an existing procedure in a new
        :class:`.Factor`, instead of constructing a new
        copy of the :class:`.Procedure`.

    :param absent:
        Whether the absence, rather than presence, of this
        :class:`.Procedure` is being referenced. The usefulness
        of this is unclear, but I'm not prepared to eliminate it
        before the :class:`.Argument` class has been implemented
        and tested.

    :param generic:
        Whether the this :class:`Procedure` is being referenced
        as a generic example, which could be replaced by any
        other :class:`Procedure`.
    """

    outputs: Tuple[Factor, ...] = ()
    inputs: Tuple[Factor, ...] = ()
    despite: Tuple[Factor, ...] = ()
    name: Optional[str] = None
    absent: bool = False
    generic: bool = False
    context_factor_names: ClassVar = ("outputs", "inputs", "despite")

    def __post_init__(self):
        outputs = self.__class__._wrap_with_tuple(self.outputs)
        inputs = self.__class__._wrap_with_tuple(self.inputs)
        despite = self.__class__._wrap_with_tuple(self.despite)
        groups = {"outputs": outputs, "inputs": inputs, "despite": despite}
        for group in groups:
            for factor_obj in groups[group]:
                if not isinstance(factor_obj, Factor):
                    raise TypeError(
                        "Input, Output, and Despite groups must contain "
                        + "only subclasses of Factor, but "
                        + f"{factor_obj} was type {type(factor_obj)}"
                    )
            object.__setattr__(self, group, groups[group])

    def __add__(self, other: Procedure) -> Optional[Procedure]:
        if not isinstance(other, self.__class__):
            return self.add_factor(other)
        matchlist = self.triggers_next_procedure(other)
        if matchlist:
            # Arbitrarily choosing the first match to decide what
            # generic Factors appear in the new outputs.
            # Wouldn't it have been better to get just one match with a generator?
            triggered_rule = other.new_context(matchlist[0])
            new_outputs = [*self.outputs, *triggered_rule.outputs]
            unique_new_outputs = tuple({key: None for key in new_outputs})
            return self.evolve({"outputs": unique_new_outputs})
        return None

    @use_likely_context
    def union(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> Optional[Procedure]:
        r"""
        Combine two :class:`Procedure`\s into one.

        The new :class:`Procedure` will have all of the ``inputs``, ``outputs``,
        and ``despite`` :class:`.Factor`\s of both ``self`` and ``other``.

        All of the context :class:`.Factor`\s of ``self`` will
        remain the same.
        """

        def combine_factor_list(
            self_list, other_list, allow_contradictions: bool = False
        ):
            new_factors = []
            for self_factor in self_list:
                broadest = self_factor
                for other_factor in other_list:
                    if allow_contradictions is False and other_factor.contradicts(
                        self_factor
                    ):
                        return None
                    if other_factor >= self_factor:
                        broadest = other_factor
                new_factors.append(broadest)
            return new_factors + [
                factor for factor in other_list if factor not in new_factors
            ]

        other = other.new_context(context)
        new_inputs = combine_factor_list(self.inputs, other.inputs)
        new_outputs = combine_factor_list(self.outputs, other.outputs)
        new_despite = combine_factor_list(
            self.despite, other.despite, allow_contradictions=True
        )
        if any(group is None for group in (new_outputs, new_inputs, new_despite)):
            return None
        return Procedure(outputs=new_outputs, inputs=new_inputs, despite=new_despite)

    def __or__(self, other):
        return self.union(other)

    def __len__(self):
        r"""
        Get number of generic :class:`.Factor`\s specified for ``self``.

        :returns:
            the number of generic :class:`.Factor`\s that need to be
            specified for this :class:`Procedure`.
        """

        return len(self.generic_factors)

    def __repr__(self):
        text = (
            f"{self.__class__.__name__}(outputs=("
            + f"{', '.join(repr(factor) for factor in self.outputs)}), "
            + f"inputs=({', '.join(repr(factor) for factor in self.inputs)}), "
            + f"despite=({', '.join(repr(factor) for factor in self.despite)}))"
        )
        return text

    def __str__(self):

        text = "RESULT:"
        for f in self.outputs:
            text += "\n" + indented(str(f))
        if self.inputs:
            text += "\nGIVEN:"
            for f in self.inputs:
                text += "\n" + indented(str(f))
        if self.despite:
            text += "\nDESPITE:"
            for f in self.despite:
                text += "\n" + indented(str(f))

        return text

    @property
    def factors_all(self) -> List[Factor]:
        r"""
        Get :class:`.Factor`\s in ``inputs``, ``outputs``, and ``despite``.

        :returns:
            a :class:`list` of all :class:`.Factor`\s.
        """

        inputs = self.inputs or ()
        despite = self.despite or ()
        return [*self.outputs, *inputs, *despite]

    @property
    def generic_factors(self) -> List[Factor]:
        r"""
        :class:`.Factor`\s that can be replaced without changing ``self``\s meaning.

        If ``self.generic is True`` then the only generic :class:`.Factor` is ``self``.
        This could happen if the :class:`.Procedure` was mentioned generically in an
        :class:`.Argument` about preserving objections for appeal, for instance.

        :returns:
            ``self``'s generic :class:`.Factor`\s,
            which must be matched to other generic :class:`.Factor`\s to
            perform equality or implication tests between :class:`.Factor`\s
            with :meth:`.Factor.means` or :meth:`.Factor.__ge__`.
        """
        if self.generic:
            return (self,)
        return list(
            {
                generic: None
                for factor in self.factors_all
                for generic in factor.generic_factors
            }
        )

    def add_factor(self, incoming: Factor, role: str = "inputs") -> Procedure:
        """
        Add an output, input, or despite :class:`.Factor`.

        :param incoming:
            the new :class:`.Factor` to be added to input, output, or despite

        :param role:
            specifies whether the new :class:`.Factor` should be added to
            input, output, or despite

        :returns:
            a new version of ``self`` with the specified change
        """

        if role not in self.context_factor_names:
            raise ValueError(f"'role' must be one of {self.context_factor_names}")
        old_factors = self.__dict__.get(role) or []
        new_factors = list(old_factors) + [incoming]
        return self.evolve({role: new_factors})

    def contradicts(self, other) -> bool:
        r"""
        Find if ``self`` applying in some cases implies ``other`` cannot apply in some.

        Raises an error because, by analogy with :meth:`.Procedure.implies`\,
        users might expect this method to return ``True`` only when
        :class:`Rule` with ``universal=False`` and Procedure ``self``
        would contradict another :class:`Rule` with ``universal=False``
        and Procedure ``other``. But that would never happen.
        """
        raise NotImplementedError(
            "Procedures do not contradict one another unless one of them ",
            "applies in 'ALL' cases. Consider using ",
            "'Procedure.contradicts_some_to_all' or 'Rule.contradicts'.",
        )

    def contradicts_some_to_all(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> bool:
        r"""
        Find if ``self`` applying in some cases implies ``other`` cannot apply in all.

        :returns:
            whether the assertion that ``self`` applies in
            **some** cases contradicts that ``other`` applies in **all**
            cases, where at least one of the :class:`.Rule`\s is ``mandatory``.
        """

        if not isinstance(other, self.__class__):
            return False
        return any(self.explain_contradiction_some_to_all(other, context))

    def explain_contradiction_some_to_all(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        """Explain why ``other`` can't apply in all cases if ``self`` applies in some."""
        if context is None:
            context = ContextRegister()
        self_despite_or_input = (*self.despite, *self.inputs)

        # For self to contradict other, every input of other
        # must be implied by some input or despite factor of self.
        relations = (Analogy(other.inputs, self_despite_or_input, operator.le),)
        matchlist = all_analogy_matches(relations, inverse=True, context=context)

        # For self to contradict other, some output of other
        # must be contradicted by some output of self.

        for m in matchlist:
            if contradictory_factor_groups(self.outputs, other.outputs, m):
                yield m

    def explain_implication_all_to_all(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:

        if isinstance(other, self.__class__):

            yield from self.explanations_same_meaning(other, context)

            relations = (
                Analogy(other.outputs, self.outputs, operator.le),
                Analogy(self.inputs, other.inputs, operator.le),
            )
            matchlist = all_analogy_matches(relations)

            # For every factor in other, find the permutations of entity slots
            # that are consistent with matchlist and that don't cause the factor
            # to contradict any factor of self.

            for matches in matchlist:
                if consistent_factor_groups(self.inputs, other.despite, matches):
                    yield matches

    def implies_all_to_all(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> bool:
        """
        Find if ``self`` applying in all cases implies ``other`` applies in all.

        ``self`` does not imply ``other`` if any output of ``other``
        is not equal to or implied by some output of ``self``.

        For ``self`` to imply ``other``, every input of ``self``
        must be implied by some input of ``other``.

        ``self`` does not imply ``other`` if any despite of ``other``
        :meth:`~.Factor.contradicts` an input of ``self``.

        :returns:
            whether the assertion that ``self`` applies in **all** cases
            implies that ``other`` applies in **all** cases.
        """
        return any(self.explain_implication_all_to_all(other, context))

    def explain_implication_all_to_some(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:

        if isinstance(other, self.__class__):

            yield from self.explain_implication_all_to_all(other)

            other_despite_or_input = (*other.despite, *other.inputs)
            self_despite_or_input = (*self.despite, *self.inputs)

            relations = (
                Analogy(other.outputs, self.outputs, operator.le),
                Analogy(other.despite, self_despite_or_input, operator.le),
            )

            matchlist = all_analogy_matches(relations)

            for context in matchlist:
                if consistent_factor_groups(
                    self.inputs, other_despite_or_input, context
                ):
                    yield context

    def implies_all_to_some(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> bool:
        r"""
        Find if ``self`` applying in all cases implies ``other`` applies in some.

        For ``self`` to imply ``other``, every output of ``other``
        must be equal to or implied by some output of ``self``.

        For ``self`` to imply ``other``, every input of ``self`` must not be
        contradicted by any input or despite of ``other``.

        ``self`` does not imply ``other`` if any "despite" :class:`.Factor`\s
        of ``other`` are not implied by inputs of ``self``.

        :returns:
            whether the assertion that ``self`` applies in **all** cases
            implies that ``other`` applies in **some** cases (that is, if
            the list of ``self``'s inputs is not considered an exhaustive list
            of the circumstances needed to invoke the procedure).
        """
        return any(self.explain_implication_all_to_some(other, context))

    def _implies_if_present(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        r"""
        Find if ``self`` would imply ``other`` if they aren't absent.

        When ``self`` and ``other`` are included in
        :class:`Rule`\s that both apply in **some** cases:

        ``self`` does not imply ``other`` if any input of ``other``
        is not equal to or implied by some input of ``self``.

        ``self`` does not imply ``other`` if any output of ``other``
        is not equal to or implied by some output of self.

        ``self`` does not imply ``other`` if any despite of ``other``
        is not equal to or implied by some despite or input of ``self``.

        :returns:
            whether the assertion that ``self`` applies in some cases
            implies that the :class:`.Procedure` ``other`` applies
            in some cases.
        """
        if isinstance(other, self.__class__):
            despite_or_input = (*self.despite, *self.inputs)

            relations = (
                Analogy(other.outputs, self.outputs, operator.le),
                Analogy(other.inputs, self.inputs, operator.le),
                Analogy(other.despite, despite_or_input, operator.le),
            )

            yield from iter(all_analogy_matches(relations, context=context))

    def explanations_same_meaning(
        self, other, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        if isinstance(other, self.__class__):
            # Verifying that every factor in self is in other.
            # Also verifying that every factor in other is in self.

            groups = ("outputs", "inputs", "despite")
            matchlist = [ContextRegister()]
            for group in groups:
                updater = Analogy(
                    need_matches=self.__dict__[group],
                    available=other.__dict__[group],
                    comparison=means,
                )
                matchlist = updater.update_matchlist(matchlist)
            if not bool(matchlist):
                return False

            # Now doing the same thing in reverse
            matchlist = [ContextRegister()]
            for group in groups:
                updater = Analogy(
                    need_matches=other.__dict__[group],
                    available=self.__dict__[group],
                    comparison=means,
                )
                matchlist = updater.update_matchlist(matchlist)
            for context in matchlist:
                yield context

    def means(self, other, context: Optional[ContextRegister] = None) -> bool:
        r"""
        Determine whether ``other`` has the same meaning as ``self``.

        :returns:
            whether the two :class:`.Procedure`\s have all the same
            :class:`.Factor`\s with the same context factors in the
            same roles.
        """
        return any(self.explanations_same_meaning(other, context))

    @new_context_helper
    def new_context(self, changes: ContextRegister) -> Procedure:
        r"""
        Create new :class:`Procedure`, replacing keys of ``changes`` with values.

        :param changes:
            a :class:`dict` of :class:`.Factor`\s to replace
            matched to the :class:`.Factor`\s that should replace them

        :returns:
            new :class:`.Procedure` object, replacing keys of
            ``changes`` with their values.
        """
        new_dict = self.__dict__.copy()
        for name in self.context_factor_names:
            new_dict[name] = tuple(
                factor.new_context(changes) for factor in new_dict[name]
            )
        return self.__class__(**new_dict)

    def triggers_next_procedure(self, other: Procedure) -> List[ContextRegister]:
        r"""
        Test if :class:`.Factor`\s from firing ``self`` would trigger ``other``.

        .. Note::
            To be confident that ``self`` actually can trigger ``other``,
            we would have to assume that ``other`` has ``universal: True``
            since otherwise we don't know exactly what is required to
            trigger ``other``.

        :param other:
            another :class:`Procedure` to test to see whether it can
            be triggered by triggering ``self``

        :returns:
            whether the set of :class:`Factor`\s that exist after ``self``
            is fired could trigger ``other``
        """
        self_despite_or_input = (*self.despite, *self.inputs)
        self_output_or_input = (*self.outputs, *self.inputs)

        relations = (
            Analogy(other.inputs, self_output_or_input, operator.le),
            Analogy(other.despite, self_despite_or_input, operator.le),
        )
        return all_analogy_matches(relations)


def consistent_factor_groups(
    self_factors: Tuple[Factor, ...],
    other_factors: Tuple[Factor, ...],
    matches: Optional[ContextRegister] = None,
):
    r"""
    Find whether two sets of :class:`.Factor`\s can be consistent.

    Works by first determining whether one :class:`.Factor`
    potentially :meth:`~.Factor.contradicts` another,
    and then determining whether it's possible to make
    context assignments match between the contradictory
    :class:`.Factor`\s.

    .. Note::
        Does ``Factor: None`` in matches always mean that
        the :class:`.Factor` can avoid being matched in a
        contradictory way?

    :returns:
        whether unassigned context factors can be assigned in such
        a way that there's no contradiction between any factor in
        ``self_factors`` and ``other_factors``, given that some
        :class:`.Factor`\s have already been assigned as
        described by ``matches``.
    """
    if matches is None:
        matches = ContextRegister()
    for self_factor in self_factors:
        for other_factor in other_factors:
            if self_factor.contradicts(other_factor):
                if all(
                    all(
                        matches.get(key) == context_register[key]
                        or matches.get(context_register[key]) == key
                        for key in self_factor.generic_factors
                    )
                    for context_register in self_factor._context_registers(
                        other_factor, means
                    )
                ):
                    return False
    return True


def contradictory_factor_groups(
    self_factors: Tuple[Factor, ...],
    other_factors: Tuple[Factor, ...],
    context: Optional[ContextRegister] = None,
) -> bool:
    r"""
    Find whether two sets of :class:`.Factor`\s can be contradictory.

    :param self_factors:
        one set of :class:`Factor`\s with consistent context factors.
        Normally collected from the ``outputs`` attribute of a :class:`Procedure`\.

    :param other_factors:
        a second set of :class:`Factor`\s with context factors that
        are internally consistent, but may not be consistent with ``self_factors``.

    :param matches:
        keys representing context :class:`.Factor`\s in ``self_factors`` and
        values representing :class:`.Factor`\s in ``other_factors``. The
        keys and values have been found in corresponding positions
        in ``self`` and ``other``.

    :returns:
        whether any :class:`.Factor` assignment can be found that
        makes a :class:`.Factor` in the output of ``other`` contradict
        a :class:`.Factor` in the output of ``self``.
    """
    if context is None:
        context = ContextRegister()
    for other_factor in other_factors:
        for self_factor in self_factors:
            if self_factor.contradicts(other_factor, context):
                return True
    return False



"""
Converting simple structured data from XML or JSON into authorityspoke objects.

These functions will usually be called by functions from the io.loaders module
after they import some data from a file.
"""
from copy import deepcopy
import datetime
from functools import partial

from bs4 import BeautifulSoup

from typing import Any, Dict, List, Iterable, Iterator
from typing import Optional, Tuple, Type, Union

from pint import UnitRegistry

from authorityspoke.decisions import Decision
from authorityspoke.codes import Code, USCCode, USConstCode, CalCode, CFRCode
from authorityspoke.enactments import Enactment
from authorityspoke.entities import Entity
from authorityspoke.evidence import Exhibit, Evidence
from authorityspoke.factors import Factor
from authorityspoke.facts import Fact
from authorityspoke.holdings import Holding
from authorityspoke.jurisdictions import Regime
from authorityspoke.opinions import Opinion, TextLinkDict
from authorityspoke.pleadings import Allegation, Pleading
from authorityspoke.predicates import Predicate
from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule

from authorityspoke.io import anchors, schemas
from authorityspoke.io.schemas import (
    RawEnactment,
    RawHolding,
    RawPredicate,
    RawFactor,
    RawDecision,
)
from authorityspoke.io.name_index import index_names


ureg = UnitRegistry()

FACTOR_SUBCLASSES = {
    class_obj.__name__: class_obj
    for class_obj in (Allegation, Entity, Exhibit, Evidence, Fact, Pleading)
}


def get_code_uri(xml, title: str) -> str:
    """
    Build a URI for the ``Code`` based on its XML metadata.

    .. note::
        This handles California state statutes only with a
        mockup, which can only refer to the Penal and Evidence
        Codes.

    :returns:
        The `United States Legislative Markup (USLM)
        <https://github.com/usgpo/uslm>`_ identifier that
        describes the document as a whole, if available in
        the XML. Otherwise returns a pseudo-USLM identifier.
    """
    if title == "Constitution of the United States":
        return "/us/const"
    if title.startswith("Title"):
        return xml.find("main").find("title")["identifier"]
    if title.startswith("California"):
        uri = "/us-ca"
        if "Penal" in title:
            return uri + "/pen"
        else:
            return uri + "/evid"
    if title.startswith("Code of Federal Regulations"):
        title_num = title.split()[-1]
        return f"/us/cfr/t{title_num}"
    raise NotImplementedError


def get_code_title(xml) -> str:
    r"""
    Provide "title" identifier for the :class:`Code` XML.

    :returns:
        the contents of an XML ``title`` element that
        describes the ``Code``, if any. Otherwise
        returns a descriptive name that may not exactly
        appear in the XML.
    """
    uslm_title = xml.find("dc:title")
    if uslm_title:
        return uslm_title.text
    cal_title = xml.h3
    if cal_title:
        code_name = cal_title.b.text.split(" - ")[0]
        return f"California {code_name}"
    cfr_title = xml.CFRGRANULE.FDSYS.CFRTITLE
    if cfr_title:
        return f"Code of Federal Regulations Title {cfr_title.text}"
    raise NotImplementedError


def read_code(xml):
    title = get_code_title(xml)
    uri = get_code_uri(xml, title)
    if uri.startswith("/us/const"):
        code_class = USConstCode
    elif uri.startswith("/us/cfr"):
        code_class = CFRCode
    elif uri.startswith("/us/"):
        code_class = USCCode
    elif uri.startswith("/us-ca"):
        code_class = CalCode
    else:
        return Code(xml, title, uri)
    return code_class(xml, title, uri)


def read_enactment(
    record: RawEnactment, code: Optional[Code] = None, regime: Optional[Regime] = None
) -> Enactment:
    r"""
    Create a new :class:`.Enactment` object using imported JSON data.

    The new :class:`.Enactment` can be composed from a :class:`.Code`
    referenced in the ``regime`` parameter.

    :param record:
        :class:`dict` with string fields from JSON for constructing
        new :class:`.Enactment`

    :param code:
        the :class:`.Code` that is the source for this
        :class:`Enactment`

    :param regime:
        the :class:`.Regime` where the :class:`.Code` that is the
        source for this :class:`Enactment` can be found, or where
        it should be added

    :returns:
        a new :class:`Enactment` object, optionally with text links.
    """
    schema = schemas.EnactmentSchema(many=False)
    schema.context["mentioned"] = index_names(record)
    schema.context["regime"] = regime
    schema.context["code"] = code
    return schema.load(deepcopy(record))


def read_enactments(
    record: List[RawEnactment],
    code: Optional[Code] = None,
    regime: Optional[Regime] = None,
) -> List[Enactment]:
    r"""
    Create a new :class:`Enactment` object using imported JSON data.

    The new :class:`Enactment` can be composed from a :class:`.Code`
    referenced in the ``regime`` parameter.

    :param record:
        sequence of :class:`dict`\s with string fields from JSON for
        constructing new :class:`.Enactment`\s

    :param code:
        a :class:`.Code` that is the source for every :class:`Enactment`
        to be loaded, if they all come from the same :class:`.Code`

    :param regime:
        the :class:`.Regime` where the :class:`.Code`\s that are the
        source for this :class:`Enactment` can be found, or where
        it should be added

    :returns:
        a list of new :class:`Enactment` objects, optionally with text links.
    """
    schema = schemas.EnactmentSchema(many=True)
    schema.context["mentioned"] = index_names(record)
    schema.context["regime"] = regime
    schema.context["code"] = code
    return schema.load(deepcopy(record))


def read_predicate(value: RawPredicate) -> Predicate:
    schema = schemas.PredicateSchema(partial=True)
    return schema.load(value)


def read_fact(record: RawFactor) -> Fact:
    r"""
    Construct a :class:`Fact` after loading a dict from JSON.

    :param record:
        parameter values to pass to :class:`.FactSchema`\.

    :returns:
        a :class:`Fact`, with optional mentioned factors
    """
    mentioned = index_names(record)
    schema = schemas.FactSchema()
    schema.context["mentioned"] = mentioned
    return schema.load(record)


def read_factor(record: RawFactor, regime: Optional[Regime] = None, **kwargs) -> Factor:
    r"""
    Turn fields from JSON into a :class:`Factor` object.

    :param record:
        parameter values to pass to schema

    :parame regime:
        to look up any :class:`.Enactment` references

    """
    schema = schemas.FactorSchema(many=False)
    schema.context["mentioned"] = index_names(record)
    schema.context["regime"] = regime
    return schema.load(record)


def read_factors(
    record: List[RawFactor], regime: Optional[Regime] = None, **kwargs
) -> Factor:
    r"""
    Turn fields from JSON into a :class:`Factor` object.

    :param record:
        parameter values to pass to schema

    :parame regime:
        to look up any :class:`.Enactment` references

    """
    schema = schemas.FactorSchema(many=True)
    schema.context["mentioned"] = index_names(record)
    schema.context["regime"] = regime
    return schema.load(record)


def read_procedure(
    record: Dict, regime: Optional[Regime] = None, many=False
) -> Procedure:
    schema = schemas.ProcedureSchema(many=many)
    schema.context["mentioned"] = index_names(record)
    schema.context["regime"] = regime
    return schema.load(record)


def read_holding(record: RawHolding, regime: Optional[Regime] = None) -> Holding:
    r"""
    Create new :class:`Holding` object from simple datatypes from JSON input.

    Will yield multiple items if ``exclusive: True`` is present in ``record``.

    :param record:
        dict of values for constructing :class:`.Holding`

    :param regime:
        Collection of :class:`.Jurisdiction`\s and corresponding
        :class:`.Code`\s for discovering :class:`.Enactment`\s to
        reference in the new :class:`Holding`.

    :param many:
        if True, record represents a list of :class:`Holding`\s rather than
        just one.

    :returns:
        New :class:`.Holding`, and an updated dictionary with mentioned
        :class:`.Factor`\s as keys and their :class:`.TextQuoteSelector`\s
        as values.
    """
    schema = schemas.HoldingSchema(many=False)
    schema.context["mentioned"] = index_names(record)
    schema.context["regime"] = regime
    return schema.load(deepcopy(record))


def read_holdings(
    record: List[RawHolding], regime: Optional[Regime] = None
) -> List[Holding]:
    r"""
    Load a list of :class:`Holdings`\s from JSON, with optional text links.

    :param record:
        a list of dicts representing holdings, in the JSON input format

    :parame regime:
        A collection of :class:`.Jurisdiction`\s and the :class:`.Code`\s
        that have been enacted in each. Used for constructing
        :class:`.Enactment`\s referenced by :class:`.Holding`\s.

    :param index_anchors:
        whether to also return an index of text links to the created object(s)

    :returns:
        a list of :class:`.Holding` objects, optionally with
        an index matching :class:`.Factor`\s to selectors.
    """
    schema = schemas.HoldingSchema(many=True)
    schema.context["mentioned"] = index_names(record)
    schema.context["regime"] = regime
    return schema.load(deepcopy(record))


def read_decision(decision_dict: RawDecision) -> Decision:
    r"""
    Create and return one or more :class:`.Opinion` objects from a dict API response.

    Relies on the JSON format from the `Caselaw Access Project
    API <https://api.case.law/v1/cases/>`_.

    This function is a more convenient way to call read_opinions with an entire
    case from the CAP API as a single parameter.

    :param decision_dict:
        A dict created from a Caselaw Access Project API response.
    """
    schema = schemas.DecisionSchema()
    return schema.load(decision_dict)


def read_rule(record: Dict, regime: Optional[Regime] = None) -> Rule:
    r"""
    Make :class:`Rule` from a :class:`dict` of fields and a :class:`.Regime`\.

    :param record:

    :param regime:

    :returns:
        iterator yielding :class:`Rule`\s with the items
        from ``mentioned_entities`` as ``context_factors``
    """
    mentioned = index_names(record)
    schema = schemas.RuleSchema()
    schema.context["mentioned"] = mentioned
    schema.context["regime"] = regime
    return schema.load(record)



"""Classes representing legislation."""

from __future__ import annotations

from typing import List, Optional, Union
from dataclasses import dataclass

from anchorpoint.textselectors import TextQuoteSelector, TextPositionSelector

from authorityspoke.codes import Code
from authorityspoke.formatting import wrapped


@dataclass(frozen=True)
class Enactment:
    """
    A passage of legislative text.

    May be used as support for a
    :class:`.Rule`. To retrieve the text, there needs
    to be an available method for identifying the correct XML
    element based on the section and subsection names, and each
    XML format used for any :class:`Code` will require a different method.

    :param selector:
        identifier for the section of a :class:`Code` where
        the text can be found.

    :param selector:
        identifier for the part of the section being cited

    :param code:
        the :class:`Code` where this legislative text appears.

    :param name:
        an identifier for this object, often used if the object needs
        to be referred to multiple times in the process of composing
        other :class:`.Factor` objects.
    """

    source: Optional[str]
    selector: Optional[Union[TextQuoteSelector, TextPositionSelector]] = None
    code: Optional[Code] = None
    name: Optional[str] = None

    def __post_init__(self):
        if not self.source:
            if not self.code:
                raise AttributeError("A Source or Code parameter is required.")
            object.__setattr__(self, "source", self.code.uri)

        if isinstance(self.selector, TextPositionSelector):
            if not self.code:
                raise AttributeError(
                    "Because text alignment may vary between Code versions, "
                    + "a Code parameter is required when the selector is a "
                    + "TextPositionSelector."
                )
            object.__setattr__(self, "_interval", self.selector)
            object.__setattr__(
                self,
                "selector",
                self._interval.as_quote_selector(
                    self.code.section_text_from_path(self.source)
                ),
            )

        if self.selector and self.code:
            text = self.code.section_text_from_path(self.source)
            if not self.selector.exact:
                object.__setattr__(
                    self, "selector", self.selector.rebuild_from_text(text=text)
                )
            interval = self.selector.as_position(text)
            if interval is None:
                raise ValueError(f"Selected text not found in source {self.source}")
            object.__setattr__(self, "_interval", interval)
        else:
            object.__setattr__(self, "_interval", None)

    def __add__(self, other):
        if other.__class__.__name__ == "Rule":
            return other + self
        if not isinstance(other, self.__class__):
            raise TypeError

        if self >= other and self.source.startswith(other.source):
            return self
        if other >= self and other.source.startswith(self.source):
            return other
        combined = self.combine_text(other) or other.combine_text(self)
        return combined

    def _combine_text(self, other: Enactment) -> Optional[Enactment]:
        r"""
        Create new :class:`Enactment` with combined text of the source :class:`Enactment`\s.

        Private method because it only works if other's path includes
        self's path.
        """
        code = self.code or other.code
        if not code:
            raise AttributeError(
                "Code attribute is needed to combine text of Enactments "
                + "with different source URIs."
            )
        section_text = code.section_text_from_path(self.source)
        self_interval = self.selector.as_position(section_text)
        other_interval = other.selector.as_position(section_text)
        new_interval = self_interval + other_interval
        if not new_interval:
            return None
        new_selector = new_interval.as_quote_selector(section_text)
        return Enactment(source=self.source, selector=new_selector, code=self.code,)

    def combine_text(self, other: Enactment) -> Optional[Enactment]:
        r"""
        Create new :class:`Enactment` with combined text of the source :class:`Enactment`\s.

        :param other:
            another :class:`Enactment` with text to combine with the text from ``self``.

        :returns:
            new :class:`Enactment` with combined text of the source :class:`Enactment`\s, or
            ``None`` if the :class:`Enactment`\s can't be combined.
        """
        if other.source == self.source:
            new_interval = self.interval + other.interval
            if new_interval is None:
                return None
            return Enactment(source=self.source, selector=new_interval, code=self.code,)
        if other.source.startswith(self.source):
            return self._combine_text(other)
        elif self.source.startswith(other.source):
            return other._combine_text(self)
        return None

    @property
    def effective_date(self):
        r"""
        Give effective date for the :class:`Enactment`\.

        :returns:
            the effective date of the text in this passage.
            Currently works only for the US Constitution.
        """
        return self.code.provision_effective_date(self.source)

    @property
    def interval(self):
        """Return :class:`TextPositionSelector` for cited part of the Enactment section."""
        return self._interval or TextPositionSelector(0, len(self.text))

    @property
    def text(self):
        r"""
        Get passage from self's Code with self's TextQuoteSelector.

        :returns: the full text of the cited passage from the XML.
        """

        return self.code.select_text(path=self.source, selector=self.selector)

    def means(self, other: Enactment) -> bool:
        r"""
        Find whether meaning of ``self`` is equivalent to that of ``other``.

        ``Self`` must be neither broader nor narrower than ``other``, which
        means it must contain the same legislative text in the same :class:`.Code`
        from the same :class:`.Jurisdiction`

        .. note::
            You could always make the result ``False`` by comparing longer
            passages of text until you found a difference between the two
            sites in the text. Does this undercut the usefulness of the
            ``means`` method?

        :returns:
            whether ``self`` and ``other`` represent the same text
            issued by the same sovereign in the same level of
            :class:`Enactment`\.
        """
        if not isinstance(other, self.__class__):
            return False

        return (
            self.text.strip(",:;. ") == other.text.strip(",:;. ")
            and self.code.jurisdiction == other.code.jurisdiction
            and self.code.level == other.code.level
        )

    def __str__(self):
        return wrapped(f'"{self.text}" ({self.code.title}, {self.source})')

    def __ge__(self, other):
        """
        Tells whether ``self`` implies ``other``.

        .. note::
            Why does this method not require the same ``code.sovereign``
            and ``code.level``, especially considering that
            :meth:`means` does?

        :returns:
            Whether ``self`` "implies" ``other``, which in this context means
            whether ``self`` contains at least all the same text as ``other``.
        """
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"{self.__class__} objects may only be compared for "
                + "implication with other Enactment objects."
            )

        return other.text.strip(",:;. ") in self.text

    def __gt__(self, other) -> bool:
        """Test whether ``self`` implies ``other`` without having same meaning."""
        if self == other:
            return False
        return self >= other


def consolidate_enactments(enactments: List[Enactment]) -> List[Enactment]:
    r"""
    Consolidate any overlapping :class:`Enactment`\s in a :class:`list`.

    :param enactments:
        a list of :class:`Enactment`\s that may include overlapping
        legislative text within the same section

    :returns:
        a list of :class:`Enactment`\s without overlapping text
    """
    consolidated: List[Enactment] = []
    while enactments:
        match_made = False
        left = enactments.pop()
        for right in enactments:
            combined = left + right
            if combined is not None:
                enactments.remove(right)
                enactments.append(combined)
                match_made = True
                break
        if match_made is False:
            consolidated.append(left)
    return consolidated



import pytest

from authorityspoke.predicates import Predicate, ureg, Q_


class TestPredicates:
    def test_predicate_with_wrong_number_of_entities(self):
        with pytest.raises(ValueError):
            _ = Predicate("{} was a motel", reciprocal=True)

    def test_predicate_with_wrong_comparison_symbol(self):
        with pytest.raises(ValueError):
            _ = Predicate(
                "the height of {} was {}",
                comparison=">>",
                quantity=Q_("160 centimeters"),
            )

    def test_convert_false_statement_about_quantity_to_obverse(self, make_predicate):
        assert make_predicate["p7_obverse"].truth is True
        assert make_predicate["p7_obverse"].quantity == ureg.Quantity(35, "foot")
        assert make_predicate["p7"].truth is True
        assert make_predicate["p7"].comparison == "<="
        assert make_predicate["p7_obverse"].comparison == "<="

    def test_quantity_type(self, make_predicate):
        assert isinstance(make_predicate["p7"].quantity, ureg.Quantity)

    def test_quantity_string(self, make_predicate):
        assert str(make_predicate["p7"].quantity) == "35 foot"

    def test_predicate_content_comparison(self, make_predicate):
        assert make_predicate["p8_exact"].content == make_predicate["p7"].content

    def test_quantity_comparison(self, make_predicate):
        assert make_predicate["p7"].quantity_comparison() == "no more than 35 foot"
        assert make_predicate["p9"].quantity_comparison() == "no more than 5 foot"
        assert make_predicate["p1"].quantity_comparison() == ""

    def test_context_slots(self, make_predicate):
        assert make_predicate["p7"].context_slots == 2

    def test_str_for_predicate_with_number_quantity(self, make_predicate):
        assert "distance between {} and {} was at least 20" in str(
            make_predicate["p8_int"]
        )
        assert "distance between {} and {} was at least 20.0" in str(
            make_predicate["p8_float"]
        )
        assert "distance between {} and {} was at least 20 foot" in str(
            make_predicate["p8"]
        )

    @pytest.mark.parametrize(
        "sentence, index, expected",
        [
            (
                "{} was names, towns, and telephone numbers of telephone subscribers",
                0,
                "{} were names, towns,",
            ),
            (
                "all of {} and {} was at the meeting",
                0,
                "all of {} and {} was at the meeting",
            ),
            (
                "all of {} and {} was at the meeting",
                1,
                "all of {} and {} were at the meeting",
            ),
        ],
    )
    def test_make_str_plural(self, sentence, index, expected):
        plural_version = Predicate.make_context_plural(sentence=sentence, index=index)
        assert plural_version.startswith(expected)

    def test_str_not_equal(self, make_predicate):
        assert "the distance between {} and {} was not equal to 35 foot" in str(
            make_predicate["p7_not_equal"]
        )

    def test_negated_method(self, make_predicate):
        assert make_predicate["p7"].negated() == make_predicate["p7_opposite"]
        assert make_predicate["p3"].negated() == make_predicate["p3_false"]

    # Equality

    def test_predicate_equality(self, make_predicate):
        assert make_predicate["p1"] == make_predicate["p1_again"]

    def test_predicate_inequality(self, make_predicate, watt_factor):
        assert make_predicate["p2"] != make_predicate["p2_reciprocal"]
        assert make_predicate["p2"] != watt_factor["f2"]

    def test_obverse_predicates_equal(self, make_predicate):
        assert make_predicate["p7"] == make_predicate["p7_obverse"]

    def test_equal_float_and_int(self, make_predicate):
        """
        These don't evaluate equal because they don't have the same
        string, but the :meth:`.Predicate.means` method considers them
        to have the same meaning.
        """
        assert not make_predicate["p8_int"] == make_predicate["p8_float"]
        assert make_predicate["p8_int"].means(make_predicate["p8_float"])

    def test_no_equality_with_inconsistent_dimensionality(self, make_predicate):
        assert make_predicate["p9"] != make_predicate["p9_acres"]

    def test_different_truth_value_prevents_equality(self, make_predicate):
        assert make_predicate["p_murder"] != make_predicate["p_murder_whether"]
        assert make_predicate["p_murder_false"] != make_predicate["p_murder_whether"]
        assert make_predicate["p_murder_false"] != make_predicate["p_murder"]

    def test_predicate_does_not_mean_fact(self, make_predicate, watt_factor):
        assert not make_predicate["p8"].means(watt_factor["f8"])

    # Implication

    def test_greater_than_because_of_quantity(self, make_predicate):
        assert make_predicate["p8_meters"] > make_predicate["p8"]
        assert make_predicate["p8_meters"] != make_predicate["p8"]

    def test_greater_float_and_int(self, make_predicate):
        assert make_predicate["p8_higher_int"] > make_predicate["p8_float"]
        assert make_predicate["p8_int"] < make_predicate["p8_higher_int"]

    def test_any_truth_value_implies_none(self, make_predicate):
        assert make_predicate["p_murder"] > make_predicate["p_murder_whether"]
        assert make_predicate["p_murder_false"] > make_predicate["p_murder_whether"]

    def test_no_implication_by_equal_quantity(self, make_predicate):
        assert not make_predicate["p_quantity=3"] > make_predicate["p_quantity>5"]

    def test_no_implication_of_equal_quantity(self, make_predicate):
        assert not make_predicate["p_quantity>5"] > make_predicate["p_quantity=3"]

    def test_no_implication_by_greater_or_equal_quantity(self, make_predicate):
        assert not make_predicate["p_quantity>=4"] > make_predicate["p_quantity>5"]

    def test_no_implication_of_greater_or_equal_quantity(self):
        less = Predicate(
            content="The number of mice was {}", comparison=">", quantity=4
        )
        more = Predicate(
            content="The number of mice was {}", comparison=">=", quantity=5
        )
        assert not less.implies(more)

    def test_equal_implies_greater_or_equal(self, make_predicate):
        assert make_predicate["p9_exact"] > make_predicate["p9"]

    def test_implication_with_not_equal(self, make_predicate):
        assert make_predicate["p7_opposite"] > make_predicate["p7_not_equal"]

    def test_no_implication_with_inconsistent_dimensionality(self, make_predicate):
        assert not make_predicate["p9"] >= make_predicate["p9_acres"]
        assert not make_predicate["p9"] <= make_predicate["p9_acres"]

    def test_implication_with_no_truth_value(self, make_predicate):
        assert not make_predicate["p2_no_truth"] > make_predicate["p2"]
        assert make_predicate["p2"] > make_predicate["p2_no_truth"]

    def test_error_predicate_imply_factor(self, make_predicate, watt_factor):
        with pytest.raises(TypeError):
            assert make_predicate["p7_true"] > (watt_factor["f7"])
        with pytest.raises(TypeError):
            assert make_predicate["p7_true"] >= (watt_factor["f7"])

    def test_predicate_implies_none(self, make_predicate):
        assert make_predicate["p7_true"] > None

    # Contradiction

    def test_predicate_contradictions(self, make_predicate):
        assert make_predicate["p7"].contradicts(make_predicate["p7_true"])
        assert not make_predicate["p1"].contradicts(make_predicate["p1_again"])
        assert not make_predicate["p3"].contradicts(make_predicate["p7"])

    def test_contradiction_by_exact(self, make_predicate):
        assert make_predicate["p8_exact"].contradicts(make_predicate["p8_less"])

    def test_contradiction_of_exact(self, make_predicate):
        assert make_predicate["p8_less"].contradicts(make_predicate["p8_exact"])

    def test_contradiction_by_equal_quantity(self, make_predicate):
        assert make_predicate["p_quantity=3"].contradicts(
            make_predicate["p_quantity>5"]
        )

    def test_contradiction_of_equal_quantity(self, make_predicate):
        assert make_predicate["p_quantity>5"].contradicts(
            make_predicate["p_quantity=3"]
        )

    def test_no_contradiction_by_greater_or_equal_quantity(self, make_predicate):
        assert not make_predicate["p_quantity>=4"].contradicts(
            make_predicate["p_quantity>5"]
        )

    def test_no_contradiction_of_greater_or_equal_quantity(self, make_predicate):
        assert not make_predicate["p_quantity>5"].contradicts(
            make_predicate["p_quantity>=4"]
        )

    def test_error_predicate_contradict_factor(self, make_predicate, watt_factor):
        with pytest.raises(TypeError):
            make_predicate["p7_true"].contradicts(watt_factor["f7"])

    def test_no_contradiction_with_no_truth_value(self, make_predicate):
        assert not make_predicate["p2_no_truth"].contradicts(make_predicate["p2"])
        assert not make_predicate["p2"].contradicts(make_predicate["p2_no_truth"])

    def test_no_contradiction_with_inconsistent_dimensionality(self, make_predicate):
        assert not make_predicate["p9"].contradicts(make_predicate["p9_acres"])
        assert not make_predicate["p9_acres"].contradicts(make_predicate["p9"])

    def test_no_contradiction_of_none(self, make_predicate):
        assert not make_predicate["p7_true"].contradicts(None)

    def test_contradiction_with_quantity(self, make_predicate):
        assert make_predicate["p8_less"].contradicts(make_predicate["p8_meters"])



import functools


def lazyprop(func):
    """
    Decorator for properties that calculate their value
    the first time and then return the cached value after that.
    See `<https://stackoverflow.com/questions/3012421/>`_
    """
    attr_name = "_lazy_" + func.__name__

    @property
    @functools.wraps(func)
    def _lazyprop(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, func(self))
        return getattr(self, attr_name)

    return _lazyprop



"""Classes representing legislation."""

from __future__ import annotations

import datetime
import functools
import re
from typing import List, Optional, Sequence, Tuple, Union

from anchorpoint.textselectors import TextQuoteSelector, TextPositionSelector
from bs4 import BeautifulSoup

from utils.roman import from_roman


class Code:
    r"""
    A code of legislation.

    Could be a constitution, code of statutes, code of regulations,
    or collection of court rules.

    Each instance of this class depends on an XML file containing
    the code, so every new XML format will require adding a method
    to this class to ingest it.

    :param xml:
        A BeautifulSoup object created by parsing the
        ``Code``\'s XML file

    :param title:
        A name for the :class:`Code` that may have been found in
        an XML ``title`` element. For larger codes like the United
        States Code, this should identify the title number, and the
        :class:`Code` object will contain text only from that title.

    :param uri:
        The `United States Legislative Markup (USLM)
        <https://github.com/usgpo/uslm>`_ identifier that
        describes the document as a whole, if available.
        Otherwise, should be a pseudo-USLM identifier.
    """

    # namespaces for legislative XML schemas
    ns = {
        "uslm": "http://xml.house.gov/schemas/uslm/1.0",
        "dc": "http://purl.org/dc/elements/1.1/",
        "xhtml": "http://www.w3.org/1999/xhtml",
    }

    def __init__(self, xml, title: str, uri: str):
        """Link an entire XML tree to the Code."""
        self.xml = xml
        self.title = title
        self.uri = uri

    @property
    def jurisdiction(self) -> str:
        """
        Get string representing the jurisdiction from within ``uri``.

        :returns:
            The abbreviation for the jurisdiction that
            enacted the ``Code``, in USLM-like format.

            e.g. ``us`` for U.S. federal laws,
            ``us-ca`` for California state laws.
        """
        return self.uri.split("/")[1]

    @property
    def level(self) -> str:
        """
        Get level of legislation for this Code, e.g. "statute".

        :returns:
            "constitution", "statute", or "regulation"
        """
        if "Constitution" in self.title:
            return "constitution"
        elif "Regulations" in self.title:
            return "regulation"
        return "statute"

    def text_interval(
        self, selector: Optional[TextQuoteSelector] = None, path: str = ""
    ) -> Optional[TextPositionSelector]:
        """
        Find integer indices for the quoted text.

        :returns:
            A :class:`TextPositionSelector` containing the lower and upper bounds of the
            text passage quoted in ``self.selector.exact`` within the
            XML section referenced in ``self.selector.path``.
        """
        sections = self.get_sections(path)
        if not sections:
            raise ValueError(f"Section {path} does not exist in code {self.title}")

        text = self.section_text(sections)
        if not selector:
            return (0, len(text))
        return selector.as_position(text)

    def provision_effective_date(self, cite: str) -> datetime.date:
        """Give effective date for a provision within the Code."""
        raise NotImplementedError

    def get_exact_from_source(
        self, source: str, selector: TextQuoteSelector
    ) -> Optional[str]:
        """
        Use ``source`` to find text for ``exact`` parameter.

        :param source:
            path to a cited section or node, which may contain subsections

        :param selector:
            selector for the cited text passage within the cited node
        """

        sections = self.get_sections(source)
        section_text = self.section_text(sections)
        return selector.select_text(section_text)

    def make_docpath(self, path: str = "") -> str:
        """Remove Code identifier from path to get a path relative to the document."""
        docpath = path or self.uri
        if not docpath.startswith(self.uri):
            return docpath  # path could be relative to the Code already
        return docpath.replace(self.uri, "")

    def _get_sections_from_relative_path(self, docpath: str) -> Optional[BeautifulSoup]:
        """
        Get sections using relative path, assuming Code follows USLM standard.
        """
        sections = self.xml.find(identifier=docpath)
        if not sections:
            return None
        passages = sections.find_all(["chapeau", "paragraph", "content"])

        return passages

    def get_sections(self, path: str = "") -> Optional[BeautifulSoup]:
        r"""
        Get sections identified by a path, if present in the :class:`Code`\.

        :param path:
            a path string, in the format used for :class:`.Enactment`
            objects, to the section with the text to be returned.

        :returns:
            the text of a section of the :class:`Code`.
        """
        docpath = self.make_docpath(path)
        if not docpath:  # selecting the whole Code
            return self.xml.find_all(name="text")
        return self._get_sections_from_relative_path(docpath)

    @staticmethod
    def section_text(passages: Sequence[BeautifulSoup]) -> str:
        """
        Get the text of legislative sections from XML elements.

        :param passages:
            a sequence of XML elements with text to join

        :returns:
            the text of the XML elements.
        """

        return " ".join(" ".join(passage.text.split()) for passage in passages)

    def section_text_from_path(self, path: str = "") -> str:
        """
        Get the text of legislative sections from a path identifier.

        :param path:
            a path string, in the format used for :class:`.Enactment`
            objects, to the section with the text to be returned.

        :returns:
            the text of the XML elements.
        """
        sections = self.get_sections(path)
        return self.section_text(sections)

    def select_text_from_interval(
        self, interval: TextPositionSelector, path: Optional[str] = None,
    ) -> Optional[str]:
        """
        Select text as interval of section identified by path.

        If ``path`` parameter is not given, selects an interval from the
        whole :class:`Code`.

        :param interval:
            the indices of the slice of the section text to be selected.

        :para path:
            a path to the section that text should be selected from

        :returns:
            a slice of text from the section identified by ``path``
        """
        sections = self.get_sections(path)
        if not sections:
            raise ValueError(f"Section {path} does not exist in {self}.")
        section_text = self.section_text(sections)
        return interval.passage(section_text)

    def select_text(
        self, path: Optional[str], selector: Optional[TextQuoteSelector]
    ) -> Optional[str]:
        r"""
        Get text from the ``Code`` using a :class:`.TextQuoteSelector`.

        :param selector:
            a selector referencing a text passage in the ``Code``.

        .. note::
            When handling Code of Federal Regulation (CFR) :class:`.Enactment`\s,
            this can only select from the whole document or from Sections,
            not Subsections or any other level. Still hoping to be able
            to switch to a `United States Legislative Markup (USLM)
            <https://github.com/usgpo/uslm>`_-like XML format for CFR.

        :returns:
            the text referenced by the selector, or ``None`` if the text
            can't be found.
        """
        sections = self.get_sections(path)
        text = self.section_text(sections)
        if not selector:
            return text
        return selector.select_text(text)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.title})"

    def __str__(self):
        return self.title


class USConstCode(Code):
    def format_uri_for_const(self, uri: str) -> str:
        """
        Convert ``uri`` to identifier for constitution sections.

        Although the US Constitution is published in a format
        described as USML, its section identifier strings differ from
        those in USC USML documents in that they skip the jurisdiction
        and code fields, skip the initial slash character,
        start with the section field, and convert all remaining
        slashes to hyphens.

        This will only remove the path to the current code if
        the current code is the same one referenced in the URI.
        So the example below assumes the current ``Code`` object
        has ``self.uri == "/us/const"``

        :param uri:
            an identifier in a format consistent with USC USML
            documents, e.g. ``/us/const/amendment/XIV/1``

        :returns:
            an identifier in a format found in the USLM version of
            the federal constitution, e.g. ``amendment-XIV-1``
        """
        return uri.replace(self.uri, "").lstrip("/").replace("/", "-")

    def get_fed_const_section(self, path: str) -> BeautifulSoup:
        """
        Get a section from a USLM identifier if ``self`` is the US Constitution.

        :param path:
            a USLM path to a section of the US Constitution, with or
            without the part that identifies the :class:`Code`, but without
            a namespace declaration.

        :returns:
            the XML section of ``self`` matching the path.
        """
        path = self.format_uri_for_const(path)
        path_parts: List[str] = []
        for _ in range(path.count("/") + 1):
            path_parts.append(path[: (path.index("/") if "/" in path else None)])
            path = path.replace("/", "-", 1)
        section = self.xml
        for part in path_parts:
            section = section.find(id=part)
        return section

    def _get_sections_from_relative_path(self, docpath: str) -> Optional[BeautifulSoup]:

        cited_section = self.get_fed_const_section(docpath)
        passages = cited_section.find_all(name="text")

        return passages

    def provision_effective_date(self, cite: str) -> datetime.date:
        """
        Give effective date for a provision within the Code.

        So far this method only covers the US Constitution and it
        assumes that the XML format is `United States Legislative
        Markup (USLM) <https://github.com/usgpo/uslm>`_.

        :param cite:
            a string or selector representing the XML element name for the
            the legislative provision within this ``Code``.

        :returns:
            the effective date of the cited provision
        """
        cite = self.format_uri_for_const(cite)
        if "amendment" not in cite.lower():
            return datetime.date(1788, 9, 13)
        roman_numeral = cite.split("-")[1]
        amendment_number = from_roman(roman_numeral)
        if amendment_number < 11:
            return datetime.date(1791, 12, 15)
        section = self.xml.find(id=cite)
        if section.name == "level":
            enactment_text = section.find("note").p.text
        else:
            enactment_text = section.parent.find("note").p.text
        month_first = re.compile(
            r"(?:Secretary of State|Administrator of General Services|certificate of the Archivist)"
            r"(?: accordingly issued a proclamation)?,? dated (\w+ \d\d?, \d{4}),"
        )
        day_first = re.compile(
            r"(?:Congress|Secretary of State),? dated the (\d\d?th of \w+, \d{4}),"
        )
        result = month_first.search(enactment_text)
        if result:
            return datetime.datetime.strptime(result.group(1), "%B %d, %Y").date()
        result = day_first.search(enactment_text)
        return datetime.datetime.strptime(result.group(1), "%dth of %B, %Y").date()


class USCCode(Code):
    def __str__(self):
        return f"USC {self.title}"

    def make_docpath(self, path: str = "") -> str:
        """
        Don't remove Code uri from path because USC uses full paths as identifiers.
        """
        return path or self.uri


class CFRCode(Code):
    def _get_sections_from_relative_path(self, docpath: str) -> Optional[BeautifulSoup]:

        section = docpath.split("/")[1].strip("s")
        citation = self.xml.find(name="SECTNO", text=f"§ {section}")
        if not citation:
            return None
        return citation.parent.find_all(name="P")


class CalCode(Code):
    def _get_sections_from_relative_path(self, docpath: str) -> Optional[BeautifulSoup]:
        def cal_href(docpath, href):
            """
            Test if XML element is labeled as the text of the section in ``docpath``.

            Uses `California statute XML format <http://leginfo.legislature.ca.gov/>`_.
            """
            section = docpath.split("/")[1].strip("s")
            return href and re.compile(
                r"^javascript:submitCodesValues\('" + section
            ).search(href)

        this_cal_section = functools.partial(cal_href, docpath)
        passages = self.xml.find(href=this_cal_section).parent.find_next_siblings(
            style="margin:0;display:inline;"
        )
        return passages



"""Create models of assertions accepted as factual by courts."""

from dataclasses import dataclass
import operator

from typing import ClassVar, Dict, Iterator, List, Optional, Sequence, Union

from authorityspoke.factors import new_context_helper
from authorityspoke.factors import Factor, ContextRegister
from authorityspoke.formatting import indented, wrapped
from authorityspoke.predicates import Predicate


@dataclass(frozen=True)
class Fact(Factor):
    r"""
    An assertion accepted as factual by a court.

    Often based on factfinding by a judge or jury.

    :param predicate:
        a natural-language clause with zero or more slots
        to insert ``context_factors`` that are typically the
        subject and objects of the clause.

    :param context_factors:
        a series of :class:`Factor` objects that fill in
        the blank spaces in the ``predicate`` statement.

    :param name:
        an identifier for this object, often used if the object needs
        to be referred to multiple times in the process of composing
        other :class:`Factor` objects.

    :param standard_of_proof:
        a descriptor for the degree of certainty associated
        with the assertion in the ``predicate``.

    :param absent:
        whether the absence, rather than the presence, of the legal
        fact described above is being asserted.

    :param generic:
        whether this object could be replaced by another generic
        object of the same class without changing the truth of the
        :class:`Rule` in which it is mentioned.

    :attr standards_of_proof:
        a tuple with every allowable name for a standard of
        proof, in order from weakest to strongest.

        .. note:
            If any courts anywhere in a legal regime disagree about the
            relative strength of the various standards of proof, or if
            any court considers the order context-specific, then this
            approach of hard-coding their names and order will have to change.
    """

    predicate: Predicate
    context_factors: Sequence[Factor] = ()
    name: Optional[str] = None
    standard_of_proof: Optional[str] = None
    absent: bool = False
    generic: bool = False
    standards_of_proof: ClassVar = (
        "scintilla of evidence",
        "substantial evidence",
        "preponderance of evidence",
        "clear and convincing",
        "beyond reasonable doubt",
    )

    def __post_init__(self):

        if (
            self.standard_of_proof
            and self.standard_of_proof not in self.standards_of_proof
        ):
            raise ValueError(
                f"standard of proof must be one of {self.standards_of_proof} or None."
            )

        if not isinstance(self.context_factors, tuple):
            context_factors = self.__class__._wrap_with_tuple(self.context_factors)
            object.__setattr__(self, "context_factors", context_factors)

        if len(self.context_factors) != len(self.predicate):
            message = (
                "The number of items in 'context_factors' must be "
                + f"{len(self.predicate)}, to match predicate.context_slots "
                + f"for '{self.predicate.content}'"
            )
            if hasattr(self, "name"):
                message += f" for '{self.name}'"
            raise ValueError(message)
        if any(not isinstance(s, Factor) for s in self.context_factors):
            raise TypeError(
                "Items in the context_factors parameter should "
                + "be Factor or a subclass of Factor, or should be integer "
                + "indices of Factor objects in the case_factors parameter."
            )

    def __str__(self):
        unwrapped = str(self.predicate.content_with_entities(self.context_factors))
        text = wrapped(super().__str__().format(unwrapped))
        if self.standard_of_proof:
            text += f"\n" + indented("by the STANDARD {self.standard_of_proof}")
        concrete_context = [
            factor for factor in self.context_factors if not factor.generic
        ]
        if any(concrete_context) and not self.generic:
            text += f"\n" + indented("SPECIFIC CONTEXT:")
            for factor in concrete_context:
                factor_text = indented(str(factor), tabs=2)
                text += f"\n{str(factor_text)}"
        return text

    @property
    def short_string(self):
        """Create one-line string representation for inclusion in other Facts."""
        predicate = str(self.predicate.content_with_entities(self.context_factors))
        standard = (
            f"by the standard {self.standard_of_proof}, "
            if self.standard_of_proof
            else ""
        )
        string = f"{standard}{predicate}"
        return super().__str__().format(string).replace("Fact", "fact")

    @property
    def interchangeable_factors(self) -> List[ContextRegister]:
        r"""
        Get ways to reorder context :class:`Factor`\s without changing truth value of ``self``.

        Each :class:`dict` returned has :class:`Factor`\s to replace as keys,
        and :class:`Factor`\s to replace them with as values.
        If there's more than one way to rearrange the context factors,
        more than one :class:`dict` should be returned.

        Currently the predicate must be phrased either in a way that
        doesn't make any context factors interchangeable, or if the
        ``reciprocal`` flag is set, in a way that allows only the
        first two context factors to switch places.

        :returns:
            the ways the context factors referenced by the
            :class:`Factor` object can be reordered without changing
            the truth value of the :class:`Factor`.

        """
        if self.predicate and self.predicate.reciprocal:
            return [
                ContextRegister(
                    {
                        self.context_factors[1]: self.context_factors[0],
                        self.context_factors[0]: self.context_factors[1],
                    }
                )
            ]
        return []

    @property
    def content(self) -> Optional[str]:
        """Access :attr:`~Predicate.content` attribute."""
        return self.predicate.content

    @property
    def truth(self) -> Optional[bool]:
        """Access :attr:`~Predicate.truth` attribute."""
        return self.predicate.truth

    def _means_if_concrete(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        if (
            isinstance(other, self.__class__)
            and self.predicate.means(other.predicate)
            and self.standard_of_proof == other.standard_of_proof
        ):
            yield from super()._means_if_concrete(other, context)

    def __len__(self):
        return len(self.context_factors)

    def _implies_if_concrete(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        """
        Test if ``self`` impliess ``other``, assuming they are not ``generic``.

        :returns:
            whether ``self`` implies ``other`` under the given assumption.
        """
        if (
            isinstance(other, self.__class__)
            and bool(self.standard_of_proof) == bool(other.standard_of_proof)
            and not (
                self.standard_of_proof
                and (
                    self.standards_of_proof.index(self.standard_of_proof)
                    < self.standards_of_proof.index(other.standard_of_proof)
                )
            )
            and self.predicate >= other.predicate
        ):
            yield from super()._implies_if_concrete(other, context)

    def _contradicts_if_present(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        """
        Test if ``self`` contradicts :class:`Fact` ``other`` if neither is ``absent``.

        :returns:
            whether ``self`` and ``other`` can't both be true at
            the same time under the given assumption.
        """
        if context is None:
            context = ContextRegister()
        if isinstance(other, Fact) and self.predicate.contradicts(other.predicate):
            yield from self._context_registers(other, operator.ge, context)

    @new_context_helper
    def new_context(self, changes: Dict[Factor, Factor]) -> Factor:
        """
        Create new :class:`Factor`, replacing keys of ``changes`` with values.

        :returns:
            a version of ``self`` with the new context.
        """
        return self.evolve(
            {
                "context_factors": tuple(
                    factor.new_context(changes) for factor in self.context_factors
                )
            }
        )


def build_fact(
    predicate: Predicate,
    indices: Optional[Sequence[int]] = None,
    case_factors: Optional[Union[Factor, Sequence[Factor]]] = None,
    name: Optional[str] = None,
    standard_of_proof: Optional[str] = None,
    absent: bool = False,
    generic: bool = False,
):
    """
    Build a :class:`.Fact` with generics selected from a list.

    :param predicate:
        a natural-language clause with zero or more slots
        to insert ``context_factors`` that are typically the
        subject and objects of the clause.

    :param context_factors:
        a series of integer indices of generic factors to
        fill in the blanks in the :class:`.Predicate`

    :param name:
        an identifier for this object, often used if the object needs
        to be referred to multiple times in the process of composing
        other :class:`.Factor` objects

    :param standard_of_proof:
        a descriptor for the degree of certainty associated
        with the assertion in the :class:`.Predicate`

    :param absent:
        whether the absence, rather than the presence, of the legal
        fact described above is being asserted.

    :param generic:
        whether this object could be replaced by another generic
        object of the same class without changing the truth of the
        :class:`Rule` in which it is mentioned.

    :param case_factors:
        a series of :class:`.Factor`\s that have already been mentioned
        in the :class:`.Opinion`. They are available for composing the
        new :class:`.Factor` object and don't need to be recreated.
    """
    if not indices:
        indices = range(len(predicate))
    if isinstance(indices, int):
        indices = (indices,)

    if case_factors is None:
        case_factors = []
    if isinstance(case_factors, Factor):
        case_factors = [case_factors]

    context_factors = tuple(case_factors[i] for i in indices)
    return Fact(
        predicate=predicate,
        context_factors=context_factors,
        name=name,
        standard_of_proof=standard_of_proof,
        absent=absent,
        generic=generic,
    )
